# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import itertools
import logging
import multiprocessing as mp

import numpy as np
import torch

from pytest import approx, mark, skip

from lenskit.parallel import invoker
from lenskit.parallel.config import _resolve_parallel_config
from lenskit.parallel.ray import ray_supported
from lenskit.testing import set_env_var

_log = logging.getLogger(__name__)


def _mul_op(m, v):
    return m @ v


@mark.slow
@mark.parametrize("pkg,n_jobs", itertools.product(["numpy", "torch"], [None, 1, 2, 4, "ray"]))
def test_invoke_matrix(pkg, n_jobs, rng: np.random.Generator):
    if n_jobs == "ray" and not ray_supported():
        skip("ray not supported")

    matrix = rng.normal(size=(1000, 1000))
    vectors = [rng.normal(size=1000) for i in range(100)]
    if pkg == "torch":
        matrix = torch.from_numpy(matrix)
        vectors = [torch.from_numpy(v) for v in vectors]
    with invoker(matrix, _mul_op, n_jobs) as inv:
        mults = inv.map(vectors)
        for rv, v in zip(mults, vectors):
            act_rv = matrix @ v
            assert act_rv == approx(rv, abs=1.0e-6)


def test_proc_count_default():
    with set_env_var("LK_NUM_PROCS", None):
        cfg = _resolve_parallel_config()
        assert cfg.processes == min(mp.cpu_count(), 4)


def test_proc_count_env():
    with set_env_var("LK_NUM_PROCS", "17"):
        cfg = _resolve_parallel_config()
        assert cfg.processes == 17
        assert cfg.subprocess().processes == 1


def test_thread_ev_list():
    with set_env_var("LK_NUM_THREADS", "4,2"):
        cfg = _resolve_parallel_config()
        assert cfg.threads == 4
        assert cfg.subprocess().threads == 2



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Item-based k-NN collaborative filtering.
"""

# pyright: basic
from __future__ import annotations

import gc
import warnings

import numpy as np
import pyarrow as pa
import scipy.sparse.linalg as spla
from pydantic import AliasChoices, BaseModel, Field, PositiveFloat, PositiveInt, field_validator
from scipy.sparse import coo_array, sparray
from typing_extensions import override

from lenskit import _accel
from lenskit.data import Dataset, FeedbackType, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.data.matrix import SparseRowArray
from lenskit.diagnostics import DataWarning
from lenskit.logging import Stopwatch, get_logger, item_progress, trace
from lenskit.logging.resource import cur_memory, max_memory
from lenskit.parallel import ensure_parallel_init
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)
MAX_BLOCKS = 1024


class ItemKNNConfig(BaseModel, extra="forbid"):
    "Configuration for :class:`ItemKNNScorer`."

    max_nbrs: PositiveInt = Field(20, validation_alias=AliasChoices("max_nbrs", "nnbrs", "k"))
    """
    The maximum number of neighbors for scoring each item.
    """
    min_nbrs: PositiveInt = 1
    """
    The minimum number of neighbors for scoring each item.
    """
    min_sim: PositiveFloat = 1.0e-6
    """
    Minimum similarity threshold for considering a neighbor.  Must be positive;
    if less than the smallest 32-bit normal (:math:`1.175 \\times 10^{-38}`), is
    clamped to that value.
    """
    save_nbrs: PositiveInt | None = None
    """
    The number of neighbors to save per item in the trained model (``None`` for
    unlimited).
    """
    feedback: FeedbackType = "explicit"
    """
    The type of input data to use (explicit or implicit).  This affects data
    pre-processing and aggregation.
    """
    block_size: int = 250
    """
    The block size for computing item similarity blocks in parallel.  Only
    affects performance, not behavior.
    """

    @field_validator("min_sim", mode="after")
    @staticmethod
    def clamp_min_sim(sim) -> float:
        return max(sim, float(np.finfo(np.float64).smallest_normal))

    @property
    def explicit(self) -> bool:
        """
        Query whether this is in explicit-feedback mode.
        """
        return self.feedback == "explicit"


class ItemKNNScorer(Component[ItemList], Trainable):
    """
    Item-item nearest-neighbor collaborative filtering feedback. This item-item
    implementation is based on the description of item-based CF by
    :cite:t:`deshpande:iknn` and hard-codes several design decisions found to
    work well in the previous Java-based LensKit code :cite:p:`lenskit-java`. In
    explicit-feedback mode, its output is equivalent to that of the Java
    version.

    .. note::

        This component must be used with queries containing the user's history,
        either directly in the input or by wiring its query input to the output of a
        user history component (e.g., :class:`~lenskit.basic.UserTrainingHistoryLookup`).

    Stability:
        Caller
    """

    config: ItemKNNConfig

    items: Vocabulary
    "Vocabulary of item IDs."
    item_means: np.ndarray[tuple[int], np.dtype[np.float32]] | None
    "Mean rating for each known item."
    item_counts: np.ndarray[tuple[int], np.dtype[np.int32]]
    "Number of saved neighbors for each item."
    sim_matrix: SparseRowArray
    "Similarity matrix (sparse CSR tensor)."

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        Train a model.

        The model-training process depends on ``save_nbrs`` and ``min_sim``, but *not* on other
        algorithm parameters.

        Args:
            ratings:
                (user,item,rating) data for computing item similarities.
        """
        if hasattr(self, "sim_matrix") and not options.retrain:
            return

        ensure_parallel_init()
        log = _log.bind(n_items=data.item_count, feedback=self.config.feedback)
        # Training proceeds in 2 steps:
        # 1. Normalize item vectors to be mean-centered and unit-normalized
        # 2. Compute similarities with pairwise dot products
        timer = Stopwatch()
        log.info("begining IKNN training")

        field = "rating" if self.config.explicit else None
        rmat = data.interactions().matrix().scipy(field, layout="coo").astype(np.float32)
        n_users, n_items = rmat.shape
        log.info(
            "[%s] made sparse matrix",
            timer,
            n_ratings=rmat.nnz,
            n_users=data.user_count,
        )

        rmat, means = self._center_ratings(log, timer, rmat)
        rmat = self._normalize_rows(log, timer, rmat)

        # convert matrix & its transpose to Arrow for Rust computation
        ui_mat = SparseRowArray.from_scipy(rmat.tocsr())
        iu_mat = SparseRowArray.from_scipy(rmat.T.tocsr())
        del rmat
        log.debug("[%s] prepared working matrices, memory use %s", timer, max_memory())

        log.info("[%s] computing similarity matrix", timer)
        with item_progress("Neighborhoods", total=n_items) as pb:
            smat = _accel.knn.compute_similarities(
                ui_mat, iu_mat, (n_users, n_items), self.config.min_sim, self.config.save_nbrs, pb
            )
        log.debug("[%s] computed, memory use %s", timer, max_memory())
        assert isinstance(smat, list)
        smat = pa.chunked_array(smat)
        smat = smat.combine_chunks()
        assert pa.types.is_large_list(smat.type)
        smat = SparseRowArray.from_array(smat)
        gc.collect()
        log.debug(
            "[%s] combined chunks, memory use %s (peak %s)", timer, cur_memory(), max_memory()
        )
        lengths = np.diff(smat.offsets)

        log.info(
            "[%s] found neighborhoods for %d of %d items",
            timer,
            np.sum(lengths > 0),
            n_items,
        )

        log.info("[%s] computed %d neighbor pairs", timer, len(smat.values))
        assert smat.offsets[-1].as_py() == len(smat.values), (
            f"{smat.offsets[-1]} != {len(smat.values)}"
        )

        self.items = data.items
        self.item_means = np.asarray(means)
        self.item_counts = np.diff(smat.offsets.to_numpy())
        self.sim_matrix = smat
        log.debug("[%s] done, memory use %s", timer, max_memory())

    def _center_ratings(self, log, timer, rmat: coo_array) -> tuple[sparray, np.ndarray | None]:
        if self.config.explicit:
            rmat = rmat.tocsc()
            counts = np.diff(rmat.indptr)
            sums = rmat.sum(axis=0)
            means = np.zeros(sums.shape, dtype=np.float32)
            # manual divide to avoid division by zero
            np.divide(sums, counts, out=means, where=counts > 0)
            rmat.data = rmat.data - np.repeat(means, counts)
            if np.allclose(rmat.data, 0.0):
                log.warning("normalized ratings are zero, centering is not recommended")
                warnings.warn(
                    "Ratings seem to have the same value, centering is not recommended.",
                    DataWarning,
                )
            log.debug("[%s] centered, memory use %s", timer, max_memory())
            return rmat, means
        else:
            return rmat, None

    def _normalize_rows(self, log, timer, rmat: sparray) -> coo_array:
        norms = spla.norm(rmat, 2, axis=0)
        # clamp small values to avoid divide by 0 (only appear when an entry is all 0)
        cmat = rmat / np.maximum(norms, np.finfo("f4").smallest_normal)
        assert cmat.shape == rmat.shape
        log.debug("[%s] normalized, memory use %s", timer, max_memory())
        return cmat

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        ensure_parallel_init()

        query = RecQuery.create(query)
        log = _log.bind(user_id=query.user_id, n_items=len(items))
        trace(log, "beginning prediction")

        ratings = query.user_items
        if ratings is None or len(ratings) == 0:
            if ratings is None:
                warnings.warn("no user history, did you omit a history component?", DataWarning)
            log.debug("user has no history, returning")
            return ItemList(items, scores=np.nan)

        # set up rating array
        # get rated item positions & limit to in-model items
        ri_nums = ratings.numbers(format="numpy", vocabulary=self.items, missing="negative")
        ri_mask = ri_nums >= 0
        ri_arr = pa.array(ri_nums, mask=~ri_mask)
        n_invalid = ri_arr.null_count
        n_valid = len(ratings) - n_invalid
        trace(log, "%d of %d rated items in model", n_valid, len(ratings))

        # convert target item information
        ti_nums = items.numbers(vocabulary=self.items, missing="negative")
        ti_mask = ti_nums >= 0
        ti_arr = pa.array(ti_nums, mask=~ti_mask)
        trace(log, "attempting to score %d of %d items", len(items) - ti_arr.null_count, len(items))

        if self.config.explicit:
            ri_vals = ratings.field("rating", "numpy")
            if ri_vals is None:
                raise RuntimeError("explicit-feedback scorer must have ratings")
            ri_vals = ri_vals.astype(np.float32, copy=True)

            # mean-center the rating array
            assert self.item_means is not None
            ri_vals[ri_mask] -= self.item_means[ri_nums[ri_mask]]
            ri_vals = pa.array(ri_vals, mask=~ri_mask)

            scores = _accel.knn.score_explicit(
                self.sim_matrix,
                ri_arr,
                ri_vals,
                ti_arr,
                self.config.max_nbrs,
                self.config.min_nbrs,
            ).to_numpy(zero_copy_only=False, writable=True)
            scores[ti_mask] += self.item_means[ti_nums[ti_mask]]

        else:
            scores = _accel.knn.score_implicit(
                self.sim_matrix, ri_arr, ti_arr, self.config.max_nbrs, self.config.min_nbrs
            ).to_numpy(zero_copy_only=False)

        log.debug(
            "scored %d items",
            int(np.isfinite(scores).sum()),
        )

        return ItemList(items, scores=scores)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Data set builder.
"""

# pyright: basic
from __future__ import annotations

import datetime as dt
import json
import warnings
from collections.abc import Mapping, Sequence
from os import PathLike
from typing import Any, Literal, TypeAlias, TypeVar, overload

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import structlog
from numpy.typing import ArrayLike, NDArray
from scipy.sparse import csr_array, sparray

from lenskit._accel import data as _data_accel
from lenskit.data.matrix import SparseRowArray
from lenskit.data.vocab import Vocabulary
from lenskit.diagnostics import DataError, DataWarning
from lenskit.logging import get_logger

from .container import DataContainer
from .dataset import Dataset
from .schema import (
    AllowableTroolean,
    AttrLayout,
    ColumnSpec,
    DataSchema,
    EntitySchema,
    RelationshipSchema,
    check_name,
    id_col_name,
    num_col_name,
)
from .types import ID, NPID, CoreID, IDSequence  # noqa: F401

_log = get_logger(__name__)

NPT = TypeVar("NPT", bound=np.generic)
NPArray1D: TypeAlias = np.ndarray[tuple[int], np.dtype[NPT]]

TableInput: TypeAlias = pd.DataFrame | pa.Table | dict[str, NDArray[Any]]
RelationshipEntities: TypeAlias = Sequence[str] | Mapping[str, str | None]

DuplicateAction: TypeAlias = Literal["update", "error", "overwrite"]
"""
Action to take on duplicate entities.
"""
MissingEntityAction: TypeAlias = Literal["insert", "filter", "error"]
"""
Action to take when a relationship references a missing entity.
"""


class DatasetBuilder:
    """
    Construct data sets from data and tables.

    Args:
        name:
            The name of the new dataset, or a data container or dataset to
            use as the basis for building a new dataset.
    """

    schema: DataSchema
    """
    The data schema assembled so far.  Do not modify this schema directly.
    """

    _log: structlog.stdlib.BoundLogger
    _tables: dict[str, pa.Table | None]
    _vocabularies: dict[str, Vocabulary]
    _rel_coords: dict[str, _data_accel.CoordinateTable | None]

    def __init__(self, name: str | DataContainer | Dataset | None = None):
        self._rel_coords = {}

        if isinstance(name, Dataset):
            name._ensure_loaded()
            name = name._data
        if isinstance(name, DataContainer):
            self.schema = name.schema.model_copy()
            self._tables = {n: t for (n, t) in name.tables.items()}
            self._vocabularies = {
                n: Vocabulary(name.tables[n].column(id_col_name(n)), name=n)
                for n in name.schema.entities.keys()
            }
            # reuse _rel_coords if we have them
            if name._rel_coords is not None:
                self._rel_coords = {
                    n: coo.copy() for n, coo in name._rel_coords.items() if coo is not None
                }
        else:
            self.schema = DataSchema(name=name, entities={"item": EntitySchema()})
            self._tables = {"item": None}
            self._vocabularies = {}

        self._log = _log.bind(ds_name=name)

    @property
    def name(self) -> str | None:
        """
        Get the dataset name.
        """
        return self.schema.name

    def entity_classes(self) -> dict[str, EntitySchema]:
        """
        Get the entity classes defined so far.
        """
        return self.schema.entities

    def relationship_classes(self) -> dict[str, RelationshipSchema]:
        """
        Get the relationship classes defined so far.
        """
        return self.schema.relationships

    def record_count(self, class_name: str) -> int:
        """
        Get the number of records for the specified entity or relationship class.
        """
        tbl = self._tables[class_name]
        if tbl is None:
            return 0
        else:
            return tbl.num_rows

    def entity_id_type(self, name: str) -> pa.DataType:
        """
        Get the PyArrow data type for an entity classes's identifiers.
        """
        tbl = self._tables[name]
        if tbl is None:
            raise ValueError(f"entity class {name} has no entities")

        return tbl.field(id_col_name(name)).type

    def add_entity_class(self, name: str) -> None:
        """
        Add an entity class to the dataset.

        Args:
            name:
                The name of the entity class.
        """
        if name in self._tables:
            _log.debug(f"class name “{name}” already in use")
            return

        check_name(name)

        self._log.debug("adding entity class", class_name=name)
        self.schema.entities[name] = EntitySchema()
        self._tables[name] = None

    def add_relationship_class(
        self,
        name: str,
        entities: RelationshipEntities,
        allow_repeats: bool = True,
        interaction: bool = False,
    ) -> None:
        """
        Add a relationship class to the dataset.  This usually doesn't need to
        be called; :meth:`add_relationships` and :meth:`add_interactions` will
        automatically add the relationship class if needed.

        As noted in :ref:`data-model`, a *relationship* records a relationship
        or interaction between two or more *entities*.  Interactions are usually
        between users and items.  The ``entities`` option to this method defines
        the names of the entity classes participating.

        .. note::

            The order of entity classes in ``entities`` matters, as the
            relationship matrix logic
            (:meth:`lenskit.data.RelationshipSet.matrix`) will default to using
            the first and last entity classes as the rows and columns of the
            matrix.

        Args:
            name:
                The name of the relationship class.
            entities:
                The entity classes participating in the relationship class.
            allow_repeats:
                Whether repeated records for the same combination of entities
                are allowed.
            interaction:
                Whether this is an interaction relationship.
        """
        if name in self._tables:
            raise ValueError(f"relationship class name “{name}” already defined")

        check_name(name)

        self._log.debug("adding relationship class", class_name=name)
        e_dict: dict[str, str | None]
        if isinstance(entities, Mapping):
            e_dict = dict(entities.items())
        else:
            e_dict = {e: None for e in entities}
        enames = list(e_dict.keys())
        if interaction and enames[-1] != "item":
            warnings.warn(
                "the last entity class for relationship class {} is {}, exepected item".format(
                    name, enames[-1]
                ),
                DataWarning,
                stacklevel=2,
            )

        self.schema.relationships[name] = RelationshipSchema(
            entities=e_dict,
            interaction=interaction,
            repeats=AllowableTroolean.ALLOWED if allow_repeats else AllowableTroolean.FORBIDDEN,
        )
        self._tables[name] = _empty_rel_table(enames)

    @overload
    def add_entities(
        self,
        cls: str,
        ids: IDSequence | pa.Array[Any] | pa.ChunkedArray[Any],
        /,
        *,
        duplicates: DuplicateAction = "error",
    ) -> None: ...
    @overload
    def add_entities(
        self, cls: str, frame: TableInput, /, *, duplicates: DuplicateAction = "error"
    ) -> None: ...
    def add_entities(
        self,
        cls: str,
        source: IDSequence | pa.Array[Any] | pa.ChunkedArray[Any] | TableInput,
        *,
        duplicates: DuplicateAction = "error",
    ) -> None:
        """
        Add entities to the data set.

        Args:
            cls:
                The name of the entity class (e.g. ``"item"``).
            source:
                The input data, as an array or list of entity IDs.

                .. note::
                    Data frame support will be added in a future version.
            duplicates:
                How to handle duplicate entity IDs.
        """
        if isinstance(source, pd.DataFrame):  # pragma: nocover
            raise NotImplementedError()
        if isinstance(source, pa.Table):  # pragma: nocover
            raise NotImplementedError()

        self._validate_entity_name(cls)

        if cls not in self.schema.entities:
            self.add_entity_class(cls)

        id_name = id_col_name(cls)
        log = self._log.bind(class_name=cls)

        # we have a sequence of IDs
        ids: pa.Array = pa.array(source)

        # figure out the new schema
        schema = pa.schema({id_name: ids.type})
        table = self._tables[cls]
        if table is not None:
            schema = pa.unify_schemas([table.schema, schema], promote_options="permissive")
            if not schema.equals(table.schema):
                log.debug("upgrading existing table schema")
                table = table.cast(schema)
        elif pa.types.is_integer(ids.type):
            self.schema.entities[cls].id_type = "int"
        elif pa.types.is_string(ids.type):
            self.schema.entities[cls].id_type = "str"
        else:
            raise TypeError(f"invalid ID type {ids.type}")

        id_field = schema.field(id_name)
        id_type: pa.DataType = id_field.type

        # de-duplicate and sort the IDs
        ids = pc.unique(ids).sort()

        n = len(ids)
        if n < len(source):
            raise DataError("cannot insert duplicate entity IDs")

        if not id_type.equals(ids.type):
            log.debug("casting IDs from type %s to %s", ids.type, id_type)
            ids = ids.cast(id_type)

        if table is not None:
            vocab = self._vocabularies[cls]
            nums = vocab.numbers(ids, format="arrow", missing="null")
            fresh_ids = pc.filter(ids, pc.invert(nums.is_valid()))
            assert len(fresh_ids) == nums.null_count
        else:
            fresh_ids = ids

        if len(fresh_ids) < n and duplicates == "error":
            n_dupes = n - len(fresh_ids)
            raise DataError(f"found {n_dupes} duplicate IDs, but re-inserts not allowed")

        log.debug(
            "adding %d new IDs (%d existing)",
            len(fresh_ids),
            0 if table is None else table.num_rows,
        )
        new_tbl = pa.table({id_name: fresh_ids})
        if table is None:
            table = new_tbl
        else:
            table = pa.concat_tables([table, new_tbl], promote_options="permissive")
        log.debug("have %d entities", table.num_rows)

        self._tables[cls] = table
        self._vocabularies[cls] = Vocabulary(table.column(id_name), name=cls, reorder=False)

    def add_relationships(
        self,
        cls: str,
        data: TableInput,
        *,
        entities: RelationshipEntities | None = None,
        missing: MissingEntityAction = "error",
        allow_repeats: bool = True,
        interaction: bool | Literal["default"] = False,
        _warning_parent: int = 0,
        remove_repeats: bool | Literal["exact"] = False,
    ) -> None:
        """
        Add relationship records to the data set.

        This method adds relationship records, provided as a Pandas data frame
        or an Arrow table, to the data set being built.  The relationships can
        be of a new class (in which case it will be created), or new
        relationship records for an existing class.

        For each entity ``E`` participating in the relationship, the table must
        have a column named ``E_id`` storing the entity IDs.

        Args:
            cls:
                The name of the interaction class (e.g. ``rating``,
                ``purchase``).
            data:
                The interaction data.
            entities:
                The entity classes involved in this interaction class.
            missing:
                What to do when interactions reference nonexisting entities; can
                be ``"error"`` or ``"insert"``.
            allow_repeats:
                Whether repeated interactions are allowed.
            interaction:
                Whether this is an interaction relationship or not; can be
                ``"default"`` to indicate this is the default interaction
                relationship.
            remove_repeats:
                If ``True``, repeated interactions will be removed. If ``"exact"``,
                duplicated interactions will be removed.
        """
        if isinstance(data, pd.DataFrame):
            table = pa.Table.from_pandas(data, preserve_index=False)
        elif isinstance(data, dict):
            table = pa.table(data)  # type: ignore
        else:
            table = data

        log = self._log.bind(class_name=cls, count=table.num_rows)

        rc_def = self.schema.relationships.get(cls, None)
        if rc_def is None:
            if entities is None:
                warnings.warn(
                    f"relationship class {cls} unknown and no entities specified,"
                    " inferring from columns",
                    DataWarning,
                    stacklevel=2 + _warning_parent,
                )
                entities = [c[:-3] for c in table.column_names if c.endswith("_id")]
            self.add_relationship_class(
                cls,
                entities,
                allow_repeats=allow_repeats,
                interaction=bool(interaction),
            )
            if interaction == "default":
                self.schema.default_interaction = cls
            rc_def = self.schema.relationships[cls]

        # FIXME: remove this segment when we can make it work

        if allow_repeats and not rc_def.repeats.is_forbidden:
            rc_def.repeats = AllowableTroolean.PRESENT

        link_id_cols = set()
        link_nums = {}
        link_mask = None
        for alias, e_type in rc_def.entities.items():
            e_type = e_type or alias
            ids = table.column(id_col_name(alias))
            if missing == "insert":
                log.debug("ensuring all entities exist")
                self.add_entities(e_type, pc.unique(ids), duplicates="update")
            e_tbl = self._tables.get(e_type, None)
            if e_tbl is None:  # pragma: nocover
                raise DataError(f"no entities of class {e_type}")

            e_nums = self._resolve_entity_ids(e_type, ids, e_tbl)
            if e_nums.null_count:
                if missing == "error":
                    raise DataError(f"{e_nums.null_count} unknown IDs for entity class {e_type}")
                assert missing == "filter"
                e_valid = e_nums.is_valid()
                if link_mask is None:
                    link_mask = e_valid
                else:
                    link_mask = pc.and_(link_mask, e_valid)
            link_nums[num_col_name(alias)] = e_nums
            link_id_cols.add(id_col_name(alias))

        new_table = pa.table(link_nums)

        for col in table.column_names:
            if col not in link_id_cols:
                new_table = new_table.append_column(col, table.column(col))
                if col not in rc_def.attributes:
                    self._validate_attribute_name(col)
                    rc_def.attributes[col] = ColumnSpec()

        if link_mask is not None:
            log.debug("filtering links to known entities")
            new_table = new_table.filter(link_mask)
        log.debug("adding %d new rows", new_table.num_rows)

        cur_table = self._tables[cls]

        # combine the tables to save them
        incoming_table = new_table
        if cur_table is not None:
            new_table = pa.concat_tables([cur_table, new_table], promote_options="permissive")

        if remove_repeats is True:
            new_table = self._remove_repeated_relationships(t=new_table, rel=rc_def)
        else:
            if remove_repeats == "exact":
                new_table = self._remove_duplicated_relationships(t=new_table)
                self._rel_coords[cls] = None

            if rc_def.repeats.is_forbidden:
                _log.debug("checking for repeated interactions")

                # FIXME: the coordinate table preservation will be out-of-order with
                # respect to sorted tables. This will not affect correctness for the
                # purposes of checking if duplicates exist, but will cause problems
                # if we count on locations for other purposes.  It also can increase
                # memory use, since the coordinate table keeps a reference to the
                # pre-sorted data.

                # get the coordinate table for the previously-existing entries
                coords = self._rel_coords.get(cls, None)
                if coords is None:
                    coords = _data_accel.CoordinateTable(len(rc_def.entities))
                    if cur_table is not None:
                        _log.debug("building coord table from previous interactions")
                        coo = cur_table.select(link_nums.keys())
                        coords.extend(coo.to_batches())
                        assert len(coo) == len(cur_table)
                    self._rel_coords[cls] = coords

                # add the new data
                coords.extend(incoming_table.select(link_nums.keys()).to_batches())
                n_dupes = len(coords) - coords.unique_count()
                _log.debug(
                    "coordinates: %d total, %d unique, %d dupe",
                    len(coords),
                    coords.unique_count(),
                    n_dupes,
                )

                if n_dupes:
                    if rc_def.repeats.is_allowed:
                        _log.debug("found %d repeat interactions", n_dupes)
                        rc_def.repeats = AllowableTroolean.PRESENT
                    else:
                        _log.error(
                            "found %d forbidden repeat interactions (of %d)",
                            n_dupes,
                            new_table.num_rows,
                        )
                        raise DataError(
                            f"repeated interactions not allowed for relationship class {cls}"
                        )

        log.debug(
            "saving new relationship table", total_rows=new_table.num_rows, schema=new_table.schema
        )
        self._tables[cls] = new_table

    def add_interactions(
        self,
        cls: str,
        data: TableInput,
        *,
        entities: RelationshipEntities | None = None,
        missing: MissingEntityAction = "error",
        allow_repeats: bool = True,
        default: bool = False,
        remove_repeats: bool | Literal["exact"] = False,
    ) -> None:
        """
        Add a interaction records to the data set.

        This method adds new interaction records, provided as a Pandas data
        frame or an Arrow table, to the data set being built.  The interactions
        can be of a new class (in which case it will be created), or new
        interactions for an existing class.

        For each entity ``E`` participating in the interaction, the table must
        have a column named ``E_id`` storing the entity IDs.

        Interactions should usually have user as the first entity and item as
        the last; the default interaction matrix logic uses the first and last
        entities as the rows and columns, respectively, of the interaction
        matrix.


        Args:
            cls:
                The name of the interaction class (e.g. ``rating``,
                ``purchase``).
            data:
                The interaction data.
            entities:
                The entity classes involved in this interaction class.
            missing:
                What to do when interactions reference nonexisting entities; can
                be ``"error"`` or ``"insert"``.
            allow_repeats:
                Whether repeated interactions are allowed.
            default:
                If ``True``, set this as the default interaction class (if the
                dataset has more than one interaction class).
            remove_repeats:
                If ``True``, repeated interactions will be removed. If ``"exact"``,
                duplicated interactions will be removed.
        """
        self.add_relationships(
            cls,
            data,
            entities=entities,
            missing=missing,
            allow_repeats=allow_repeats,
            interaction="default" if default else True,
            _warning_parent=1,
            remove_repeats=remove_repeats,
        )

    def filter_interactions(
        self,
        cls: str,
        min_time: int | float | dt.datetime | None = None,
        max_time: int | float | dt.datetime | None = None,
        remove: pa.Table | dict[str, ArrayLike] | pd.DataFrame | None = None,
    ):
        """
        Filter interactions based on timestamp or to remove particular entities.

        Args:
            cls:
                The interaction class to filter.
            min_time:
                The minimum interaction time to keep (inclusive).
            max_time:
                The maximum interaction time to keep (exclusive).
            remove:
                Combinations of entity numbers or IDs to remove.  The entities
                are filtered using an anti-join with this table, so providing a
                single column of entity IDs or numbers will remove all
                interactions associated with the listed entities.
        """
        tbl = self._tables[cls]
        if tbl is None:  # pragma: nocover
            raise ValueError(f"interaction class {cls} is empty")

        if min_time is not None or max_time is not None:
            if "timestamp" not in tbl.column_names:
                raise RuntimeError("timestamp column required to filter by timestamp")
            schema = tbl.schema
            ts_field = schema.field("timestamp")

        mask = None
        if min_time is not None:
            min_time = _conform_time(min_time, ts_field.type)
            mask = pc.greater_equal(tbl.column("timestamp"), min_time)
        if max_time is not None:
            max_time = _conform_time(max_time, ts_field.type)
            mask2 = pc.less(tbl.column("timestamp"), max_time)

            if mask is None:
                mask = mask2
            else:
                mask = pc.and_(mask, mask2)

        if mask is not None:
            tbl = tbl.filter(mask)

        if remove is not None:
            if isinstance(remove, pd.DataFrame):
                remove = pa.Table.from_pandas(remove, preserve_index=False)
            else:
                remove = pa.table(remove)  # type: ignore
            assert isinstance(remove, pa.Table)
            rtbl_cols = {}
            for cname in remove.column_names:
                if cname.endswith("_id"):
                    ent = cname[:-3]
                    num_col = num_col_name(ent)
                    etbl = self._tables[ent]
                    assert etbl is not None
                    col = remove.column(cname)
                    id_col = etbl.column(cname)
                    nums = pc.index_in(col, id_col)
                    rtbl_cols[num_col] = nums
                elif cname.endswith("_num"):
                    rtbl_cols[cname] = remove.column(cname)
                else:  # pragma: nocover
                    raise ValueError(f"invalid removal column {cname}")

            remove = pa.table(rtbl_cols)
            tbl = tbl.join(remove, remove.column_names, join_type="left anti")

        self._tables[cls] = tbl
        self._rel_coords[cls] = None

    def binarize_ratings(
        self,
        cls: str = "rating",
        min_pos_rating: float = 3.0,
        method: str = "remove",
    ):
        """
        Binarize the ratings in a relationship class.

        Args:
            cls : The relationship class to binarize (default: "rating").
            min_pos_rating : Minimum rating to consider as positive.
            method: 'zero' to set ratings to 0/1, 'remove' to drop rows below min_rating.
        """
        tbl = self._tables.get(cls, None)
        if tbl is None:
            raise ValueError(f"relationship class {cls} is empty")

        if "rating" not in tbl.column_names:
            raise ValueError(f"relationship class {cls} does not have a 'rating' column")

        min_rating = pa.scalar(min_pos_rating, tbl.column("rating").type)
        rating_col = tbl.column("rating")
        if not (
            pc.min(tbl.column("rating")).as_py()
            <= min_pos_rating
            <= pc.max(tbl.column("rating")).as_py()
        ):
            raise ValueError(
                f"min_pos_rating {min_pos_rating} is not in the range of ratings "
                f"[{pc.min(tbl.column('rating')).as_py()}, {pc.max(tbl.column('rating')).as_py()}]"
            )

        if method == "zero":
            # Set rating to 1 if >= min_rating, else 0
            mask = pc.greater_equal(tbl.column("rating"), min_rating)
            binarized = pc.cast(mask, rating_col.type)
            tbl = tbl.set_column(tbl.schema.get_field_index("rating"), "rating", binarized)
        elif method == "remove":
            # Keep only rows where rating >= min_rating
            mask = pc.greater_equal(tbl.column("rating"), min_rating)
            tbl = tbl.filter(mask)
        else:
            raise ValueError("method must be 'zero' or 'remove'")

        self._tables[cls] = tbl
        self._rel_coords[cls] = None

    def clear_relationships(self, cls: str):
        """
        Remove all records for a specified relationship class.
        """
        self._tables[cls] = _empty_rel_table(self.schema.relationships[cls].entity_class_names)
        self._rel_coords[cls] = None

    @overload
    def add_scalar_attribute(
        self,
        cls: str,
        name: str,
        data: pd.Series[Any] | TableInput,
        /,
        *,
        dictionary: bool = False,
    ) -> None: ...
    @overload
    def add_scalar_attribute(
        self,
        cls: str,
        name: str,
        entities: IDSequence | tuple[IDSequence, ...],
        values: ArrayLike,
        /,
        *,
        dictionary: bool = False,
    ) -> None: ...
    def add_scalar_attribute(
        self,
        cls: str,
        name: str,
        entities: IDSequence | tuple[IDSequence, ...] | pd.Series[Any] | TableInput,
        values: ArrayLike | None = None,
        *,
        dictionary: bool = False,
    ) -> None:
        """
        Add a scalar attribute to an entity class.

        Args:
            cls:
                The entity class name.
            name:
                The attribute name.
            entities:
                The IDs for the entities whose attribute should be set.
            values:
                The attribute values.
            data:
                A Pandas datatframe or Arrow table storing entity IDs and
                attribute values.
            dictionary:
                ``True`` to dictionary-encode the attribute values (saves space
                for string categorical values).
        """
        if name in self.schema.entities[cls].attributes:  # pragma: nocover
            raise NotImplementedError("updating or replacing existing attributes not supported")
        self._validate_attribute_name(name)

        id_col = id_col_name(cls)

        if values is None:
            if isinstance(entities, pd.Series):
                # it is a series, use the index
                values = entities.values
                entities = entities.index.values
            elif isinstance(entities, pd.DataFrame):
                values = entities[name].values
                if id_col in entities.columns:
                    entities = entities[id_col].values
                else:
                    entities = entities.index.values

        e_tbl = self._tables[cls]
        if e_tbl is None:  # pragma: nocover
            raise DataError(f"no entities of class {cls}")

        nums = self._resolve_entity_ids(cls, entities, e_tbl)
        if not np.all(nums.is_valid()):  # pragma: nocover
            n_bad = nums.is_valid().sum().as_py()
            raise DataError(f"{n_bad} unknown entity IDs")

        val_array: pa.Array = pa.array(values)  # type: ignore
        nums_np = nums.to_numpy()
        vals_np = np.full(e_tbl.num_rows, None, dtype=object)
        vals_np[nums_np] = val_array.to_numpy(zero_copy_only=False)
        val_col = pa.array(vals_np, type=val_array.type)

        if dictionary:
            val_col = pc.dictionary_encode(val_col)

        self._tables[cls] = e_tbl.append_column(name, val_col)
        self.schema.entities[cls].attributes[name] = ColumnSpec(layout=AttrLayout.SCALAR)

    @overload
    def add_list_attribute(
        self,
        cls: str,
        name: str,
        data: pd.Series[Any] | TableInput,
        /,
        *,
        dictionary: bool = False,
    ) -> None: ...
    @overload
    def add_list_attribute(
        self,
        cls: str,
        name: str,
        entities: IDSequence | tuple[IDSequence, ...],
        values: ArrayLike,
        /,
        *,
        dictionary: bool = False,
    ) -> None: ...
    def add_list_attribute(
        self,
        cls: str,
        name: str,
        entities: IDSequence | tuple[IDSequence, ...] | pd.Series[Any] | TableInput,
        values: ArrayLike | None = None,
        *,
        dictionary: bool = False,
    ) -> None:
        """
        Add a list attribute to an entity class.

        Args:
            cls:
                The entity class name.
            name:
                The attribute name.
            entities:
                The IDs for the entities whose attribute should be set.
            values:
                The attribute values (an array or list of lists)
            data:
                A Pandas datatframe or Arrow table storing entity IDs and
                attribute values.
            dictionary:
                ``True`` to dictionary-encode the attribute values (saves space
                for string categorical values).
        """
        if name in self.schema.entities[cls].attributes:  # pragma: nocover
            raise NotImplementedError("updating or replacing existing attributes not supported")
        self._validate_attribute_name(name)

        id_col = id_col_name(cls)

        if values is None:
            if isinstance(entities, pd.Series):
                # it is a series, use the index
                values = entities.values
                entities = entities.index.values
            elif isinstance(entities, pd.DataFrame):
                values = entities[name].values
                if id_col in entities.columns:
                    entities = entities[id_col].values
                else:
                    entities = entities.index.values

        e_tbl = self._tables[cls]
        if e_tbl is None:  # pragma: nocover
            raise DataError(f"no entities of class {cls}")
        nums = self._resolve_entity_ids(cls, entities, e_tbl)
        if not np.all(nums.is_valid()):  # pragma: nocover
            n_bad = nums.is_valid().sum().as_py()
            raise DataError(f"{n_bad} unknown entity IDs")

        val_array: pa.Array = pa.array(values)  # type: ignore
        if not pa.types.is_list(val_array.type):
            raise TypeError("attribute data did not resolve to list")

        if isinstance(val_array, pa.ChunkedArray):  # pragma: nocover
            val_array = val_array.combine_chunks()
        assert isinstance(val_array, pa.ListArray)

        if dictionary:
            val_array = pa.ListArray.from_arrays(
                val_array.offsets, pc.dictionary_encode(val_array.values)
            )

        nums = nums.to_numpy()

        # we have to do surgery on the offsets and values
        val_col = _expand_and_align_list_array(e_tbl.num_rows, nums, val_array)

        self._tables[cls] = e_tbl.append_column(name, val_col)
        self.schema.entities[cls].attributes[name] = ColumnSpec(layout=AttrLayout.LIST)

    def add_vector_attribute(
        self,
        cls: str,
        name: str,
        entities: IDSequence | tuple[IDSequence, ...],
        values: pa.Array[Any] | pa.ChunkedArray[Any] | np.ndarray[tuple[int, int], Any] | sparray,
        /,
        dim_names: ArrayLike | pd.Index[Any] | Sequence[Any] | None = None,
    ) -> None:
        """
        Add a vector attribute to a set of entities.

        .. warning::

            Dense vector attributes are stored densely, even for entities for
            which it is not set. High-dimensional vectors can therefore take up
            a lot of space.

        Args:
            cls:
                The entity class name.
            name:
                The attribute name.
            entities:
                The entity IDs to which the attribute should be attached.
            values:
                The attribute values, as a fixed-length list array or a
                two-dimensional NumPy array (for dense vector attributes) or a
                SciPy sparse array (for sparse vector attributes).
            dim_names:
                The names for the dimensions of the array.
        """
        if name in self.schema.entities[cls].attributes:  # pragma: nocover
            raise NotImplementedError("updating or replacing existing attributes not supported")
        self._validate_attribute_name(name)

        e_tbl = self._tables[cls]
        if e_tbl is None:  # pragma: nocover
            raise DataError(f"no entities of class {cls}")
        nums = self._resolve_entity_ids(cls, entities, e_tbl)
        if not np.all(nums.is_valid()):  # pragma: nocover
            n_bad = nums.is_valid().sum().as_py()
            raise DataError(f"{n_bad} unknown entity IDs")

        tbl_valid = np.zeros(e_tbl.num_rows, dtype=np.bool_)
        tbl_valid[nums.to_numpy()] = True

        metadata = {}
        if dim_names is not None:
            metadata["lenskit:names"] = json.dumps(np.asarray(dim_names).tolist())

        if isinstance(values, sparray):
            vec_col = self._add_sparse_vector_attribute(cls, name, values, nums, e_tbl, tbl_valid)
        elif isinstance(values, np.ndarray):
            vec_col = self._add_dense_vector_attribute_numpy(
                cls, name, values, nums, e_tbl, tbl_valid
            )
        else:
            vec_col = self._add_dense_vector_attribute(cls, name, values, nums, e_tbl, tbl_valid)

        field = pa.field(name, vec_col.type, nullable=True, metadata=metadata)
        self._tables[cls] = e_tbl.append_column(field, vec_col)

    def _validate_attribute_name(self, attribute_name: str):
        if attribute_name.endswith(("_id", "_num")) or attribute_name.startswith("_"):
            raise ValueError(f"invalid attribute name {attribute_name}")

    def _add_sparse_vector_attribute(
        self,
        cls: str,
        name: str,
        values: sparray,
        rows: pa.Int32Array,
        table: pa.Table,
        valid: NDArray[np.bool_],
    ):
        csr: csr_array = values.tocsr()  # type: ignore
        assert isinstance(csr, csr_array)
        csr.sort_indices()

        array = SparseRowArray.from_scipy(csr)
        vec_col = _expand_and_align_list_array(table.num_rows, rows.to_numpy(), array.storage)
        vec_col = SparseRowArray.from_array(vec_col)
        self.schema.entities[cls].attributes[name] = ColumnSpec(
            layout=AttrLayout.SPARSE, vector_size=array.dimension
        )
        return vec_col

    def _add_dense_vector_attribute(
        self,
        cls: str,
        name: str,
        values: pa.Array[Any] | pa.ChunkedArray[Any],
        rows: pa.Int32Array,
        table: pa.Table,
        valid: NDArray[np.bool_],
    ):
        if isinstance(values, pa.ChunkedArray):
            values = values.combine_chunks()
        if not pa.types.is_fixed_size_list(values.type):
            raise TypeError("attribute data must be fixed-size list")
        assert isinstance(values, pa.FixedSizeListArray)
        v_valid = values.is_valid().to_numpy(zero_copy_only=False)

        # no nulls: use as-is
        if np.all(valid) and np.all(v_valid):
            return values

        # find the rows where we have a valid column value
        c_valid = valid.copy()
        c_valid[rows] &= v_valid

        # nulls: we actually need a list array for sparsity + storage
        size = values.type.list_size
        values = values.cast(pa.list_(values.type.value_type))
        col = _expand_and_align_list_array(table.num_rows, rows.to_numpy(), values)

        # sizes = np.zeros(table.num_rows + 1, dtype=np.int32)
        # sizes[1:][c_valid] = size
        # offsets = np.cumsum(sizes, dtype=np.int32)
        # assert offsets[-1] == size * np.sum(c_valid)

        # # We need to reorder to match the order in the original table.  The
        # # argsort gives give us the order in which we need to consult rows in
        # # the matrix, but realigned to match the order in the table, not the
        # # order in the source vector.
        # ipos = np.argsort(rows)
        # if np.any(np.diff(ipos) < 0):
        #     values = values.take(ipos)

        # vfilt = values.filter(v_valid)
        # assert len(vfilt) == np.sum(c_valid)

        # col = pa.ListArray.from_arrays(pa.array(offsets), vfilt.values, mask=pa.array(~c_valid))
        assert np.all(col.is_valid().to_numpy(zero_copy_only=False) == c_valid)
        self.schema.entities[cls].attributes[name] = ColumnSpec(
            layout=AttrLayout.VECTOR, vector_size=size
        )
        return col

    def _add_dense_vector_attribute_numpy(
        self,
        cls: str,
        name: str,
        values: NDArray[Any],
        rows: pa.Int32Array,
        table: pa.Table,
        valid: NDArray[np.bool_],
    ):
        if len(values.shape) != 2:
            raise TypeError("values must be 2D array")
        nrow, ncol = values.shape
        arr = pa.FixedSizeListArray.from_arrays(values.ravel(), ncol)
        return self._add_dense_vector_attribute(cls, name, arr, rows, table, valid)

    def build(self) -> Dataset:
        """
        Build the dataset.
        """
        return Dataset(self.build_container())

    def build_container(self) -> DataContainer:
        """
        Build a data container (backing store for a dataset).
        """
        log = _log.bind(name=self.name)
        log.debug("assembling data container")
        tables = {}
        for n, t in self._tables.items():
            if t is None:
                log.debug("fabricating empty table %s", n)
                tables[n] = pa.table({id_col_name(n): pa.array([], type=pa.int64())})
            else:
                rel = self.schema.relationships.get(n, None)
                if rel is not None:
                    # currently not used due to coordinate table
                    # self._check_repeat_interactions(t, rel)
                    if rel.repeats.is_forbidden and len(rel.entities) == 2:
                        e_cols = [e + "_num" for e in rel.entities.keys()]
                        if not _data_accel.is_sorted_coo(t.to_batches(), *e_cols):
                            log.debug("sorting non-repeating relationship %s", n)
                            t = t.sort_by([(c, "ascending") for c in e_cols])

                tables[n] = t

        return DataContainer(self.schema.model_copy(), tables, _rel_coords=self._rel_coords)

    def save(self, path: str | PathLike[str]):
        """
        Save the dataset to disk in the LensKit native format.

        Args:
            path:
                The path where the dataset will be saved (will be created as a
                directory).
        """
        container = self.build_container()
        container.save(path)

    def _remove_duplicated_relationships(self, t: pa.Table) -> pa.Table:
        temp_df = t.to_pandas()
        temp_df.drop_duplicates(inplace=True)
        t = pa.Table.from_pandas(temp_df, preserve_index=False)
        return t

    def _remove_repeated_relationships(self, t: pa.Table, rel: RelationshipSchema) -> pa.Table:
        if "timestamp" in t.column_names:
            t = t.sort_by([("timestamp", "ascending")])
        t_modified = t.add_column(0, "_row_number", pa.array(np.arange(t.num_rows)))
        t_modified = t_modified.group_by(
            [entity + "_num" for entity in rel.entity_class_names]
        ).aggregate([("_row_number", "max")])
        t_modified = t_modified.sort_by([("_row_number_max", "ascending")])
        t = t.take(t_modified.column("_row_number_max"))
        return t

    def _validate_entity_name(self, entity_name: str):
        if entity_name.startswith("_"):
            raise ValueError(f"invalid entity name {entity_name}")

    def _resolve_entity_ids(
        self, cls: str, ids: IDSequence, table: pa.Table | None = None
    ) -> pa.Int32Array:
        if isinstance(ids, pa.ChunkedArray):
            tgt_ids = ids.combine_chunks()
        else:
            tgt_ids = pa.array(ids)  # type: ignore
        vocab = self._vocabularies.get(cls, None)
        if vocab is None:
            return pa.nulls(len(tgt_ids), type=pa.int32())
        return vocab.numbers(tgt_ids, format="arrow", missing="null")


def _expand_and_align_list_array(
    out_len: int, rows: NDArray[np.int32], lists: pa.ListArray
) -> pa.ListArray:
    """
    Expand a list array, so that its lists are on the specified rows of the
    output array, with other entries null.
    """
    assert pc.count_distinct(rows).as_py() == len(rows)
    assert len(lists) == len(rows)
    # only valid inputs
    valid = lists.is_valid().to_numpy(zero_copy_only=False)
    if not np.all(valid):
        rows = rows[valid]
        lists = lists.drop_null()
        assert len(lists) == len(rows)

    # reorder input to align with the output
    order = np.argsort(rows)
    if np.any(np.diff(order) < 0):
        lists = lists.take(order)
        rows = rows[order]

    # now we do surgery — put the lengths into place, and set up offsets
    sizes = np.zeros(out_len + 1, dtype=np.int32)
    sizes[rows + 1] = lists.value_lengths().to_numpy()
    offsets = np.cumsum(sizes, dtype=np.int32)
    offsets = pa.array(offsets)

    mask = np.ones(out_len, dtype=np.bool_)
    mask[rows] = False
    mask = pa.array(mask)

    # we can now construct the array — these offsets point into the source array
    return pa.ListArray.from_arrays(offsets, lists.values, mask=mask)


def _empty_rel_table(types: list[str]) -> pa.Table:
    return pa.table({num_col_name(t): pa.array([], pa.int32()) for t in types})


def _conform_time(time: int | float | str | dt.datetime, col_type: pa.DataType):
    if isinstance(time, str):
        time = dt.datetime.fromisoformat(time)

    if pa.types.is_timestamp(col_type):
        if not isinstance(time, dt.datetime):
            return dt.datetime.fromtimestamp(time)
    elif isinstance(time, dt.datetime):
        return time.timestamp()

    return time



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
import numpy as np
import pyarrow as pa

from pytest import raises

from lenskit.data import DatasetBuilder
from lenskit.diagnostics import DataError


def test_empty_builder():
    dsb = DatasetBuilder()
    assert not dsb.relationship_classes()
    ecs = dsb.entity_classes()
    assert len(ecs) == 1
    items = ecs["item"]
    assert items is not None
    assert items.id_type is None
    assert not items.attributes
    assert dsb.record_count("item") == 0
    with raises(KeyError):
        dsb.record_count("user")


def test_add_entity_ids():
    dsb = DatasetBuilder()

    dsb.add_entities("item", ["a", "b", "x", "y", "z"])
    icls = dsb.entity_classes()["item"]
    assert icls.id_type == "str"
    assert dsb.entity_id_type("item") == pa.string()

    ds = dsb.build()
    assert ds.item_count == 5
    assert np.all(ds.items.ids() == ["a", "b", "x", "y", "z"])
    assert np.all(np.asarray(ds.entities("item").ids()) == ["a", "b", "x", "y", "z"])


def test_add_new_entity_ids():
    dsb = DatasetBuilder()

    dsb.add_entities("user", ["a", "b", "x", "y", "z"])
    ecs = dsb.entity_classes()
    assert set(ecs.keys()) == {"user", "item"}

    ucls = ecs["user"]
    assert ucls.id_type == "str"
    assert dsb.entity_id_type("user") == pa.string()

    ds = dsb.build()
    assert ds.item_count == 0
    assert ds.user_count == 5
    assert np.all(ds.users.ids() == ["a", "b", "x", "y", "z"])
    assert np.all(ds.entities("user").ids() == ["a", "b", "x", "y", "z"])


def test_add_invalid_entity_class_name():
    dsb = DatasetBuilder()
    with raises(ValueError, match="invalid"):
        dsb.add_entities("_item", ["a", "b", "c"])

    dsb.add_entities("item", ["a", "b", "c"])
    ds = dsb.build()
    assert ds.item_count == 3
    assert np.all(ds.items.ids() == ["a", "b", "c"])


def test_add_duplicate_entities_forbidden():
    dsb = DatasetBuilder()

    dsb.add_entities("item", ["a", "b", "c"])
    with raises(DataError, match="duplicate"):
        dsb.add_entities("item", ["d", "b", "e"])


def test_add_duplicate_entities_overwrite():
    dsb = DatasetBuilder()

    dsb.add_entities("item", ["a", "b", "c"])
    dsb.add_entities("item", ["d", "b", "e"], duplicates="overwrite")

    ds = dsb.build()
    assert ds.item_count == 5
    assert set(ds.items.ids()) == {"a", "b", "c", "d", "e"}


def test_add_entities_upcast_existing():
    dsb = DatasetBuilder()

    dsb.add_entities("item", np.arange(10, dtype="i4"))
    assert dsb.entity_id_type("item") == pa.int32()

    dsb.add_entities("item", np.arange(100, 110, dtype="i8"))
    assert dsb.entity_id_type("item") == pa.int64()

    ds = dsb.build()
    assert ds.items.ids().dtype == np.int64


def test_add_entities_upcast_new():
    dsb = DatasetBuilder()

    dsb.add_entities("item", np.arange(10, dtype="i8"))
    assert dsb.entity_id_type("item") == pa.int64()

    dsb.add_entities("item", np.arange(100, 110, dtype="i4"))
    assert dsb.entity_id_type("item") == pa.int64()

    ds = dsb.build()
    assert ds.items.ids().dtype == np.int64


def test_reject_invalid_entity_id_type():
    dsb = DatasetBuilder()

    with raises(TypeError):
        dsb.add_entities("item", np.random.randn(10))  # type: ignore


def test_reject_duplicate_ids():
    dsb = DatasetBuilder()

    with raises(DataError):
        dsb.add_entities("item", ["a", "b", "a"])


def test_add_entities_twice():
    dsb = DatasetBuilder()

    dsb.add_entities("user", ["a", "b", "x", "y", "z"])
    ecs = dsb.entity_classes()
    assert set(ecs.keys()) == {"user", "item"}

    dsb.add_entities("user", ["q", "r", "s"])

    ds = dsb.build()
    assert ds.item_count == 0
    assert ds.user_count == 8
    assert np.all(ds.users.ids() == ["a", "b", "x", "y", "z", "q", "r", "s"])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Primary item-list abstraction.
"""

from __future__ import annotations

import io
import warnings
from collections.abc import Sequence

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import torch
from numpy.typing import ArrayLike, NDArray
from typing_extensions import (
    Any,
    Literal,
    LiteralString,
    overload,
)

from lenskit._accel import data as _data_accel
from lenskit.diagnostics import DataWarning
from lenskit.stats import argtopn

from .arrow import get_indexer
from .checks import check_1d
from .mtarray import MTArray, MTGenericArray
from .types import IDArray, IDSequence, NPVector
from .vocab import Vocabulary

ILIndexer = (
    np.ndarray[tuple[int], np.dtype[np.bool_]]
    | NPVector[np.integer]
    | Sequence[int]
    | torch.Tensor
    | pa.BooleanArray
    | pa.Int32Array
    | int
    | slice
)


class ItemList:
    """
    Representation of a (usually ordered) list of items, possibly with scores
    and other associated data; many components take and return item lists.  Item
    lists are to be treated as **immutable** — create a new list with modified
    data, do not do in-place modifications of the list itself or the arrays or
    data frame it returns.

    To get the length of an item list, use the standard :func:`len` function.

    An item list logically a list of rows, each of which is an item with
    multiple fields.  A designated field, ``score``, is available through the
    :meth:`scores` method, and is always single-precision floating-point.

    Item lists can be subset as an array (e.g. ``items[selector]``), where
    integer indices (or arrays thereof), boolean arrays, and slices are allowed
    as selectors.

    When an item list is pickled, it is pickled compactly but only for CPUs: the
    vocabulary is dropped (after ensuring both IDs and numbers are computed),
    and all arrays are pickled as NumPy arrays.  This makes item lists compact
    to serialize and transmit, but does mean that that serializing an item list
    whose scores are still on the GPU will deserialize on the CPU in the
    receiving process.  This is usually not a problem, because item lists are
    typically used for small lists of items, not large data structures that need
    to remain in shared memory.

    In a few places, the “items” in an item list may be other entities, such as
    users, tags, authors, etc.  This seems less confusing than calling it
    ``EntityList``, but having a very different use case and feature set than
    :class:`~lenskit.data.EntitySet`.

    .. note::

        Naming for fields and accessor methods is tricky, because the usual
        convention for a data frame is to use singular column names (e.g.
        “item_id”, “score”) instead of plural (“item_ids”, “scores”) — the data
        frame, like a database table, is a list of instances, and the column
        names are best interpreted as naming attributes of individual instances.

        However, when working with a list of e.g. item IDs, it is more natural —
        at least to this author — to use plural names: ``item_ids``.  Since this
        class is doing somewhat double-duty, representing a list of items along
        with associated data, as well as a data frame of columns representing
        items, the appropriate naming is not entirely clear.  The naming
        convention in this class is therefore as follows:

        * Field names are singular (``item_id``, ``score``).
        * Named accessor methods are plural (:meth:`item_ids`, :meth:`scores`).
        * Both singular and plural forms are accepted for item IDs numbers, and
          scores in the keyword arguments.  Other field names should be
          singular.

    .. todo::

        Right now, selection / subsetting only happens on the CPU, and will move
        data to the CPU for the subsetting operation.  There is no reason, in
        principle, why we cannot subset on GPU.  Future revisions may add
        support for this.

    Args:
        source:
            A source item list. If provided and an :class:`ItemList`, its fields
            and data are used to initialize any aspects of the item list that
            are not provided in the other arguments.  Otherwise, it is
            interpreted as ``item_ids``.
        item_ids:
            A list or array of item identifiers. ``item_id`` is accepted as an
            alternate name.
        item_nums:
            A list or array of item numbers. ``item_num`` is accepted as an
            alternate name.
        vocabulary:
            A vocabulary to translate between item IDs and numbers.
        ordered:
            Whether the list has a meaningful order.
        scores:
            An array of scores for the items.  Pass the value ``False`` to
            remove the scores when copying from a source list.
        fields:
            Additional fields, such as ``score`` or ``rating``.  Field names
            should generally be singular; the named keyword arguments and
            accessor methods are plural for readability (“get the list of item
            IDs”).  Pass the value ``False`` to remove the field when copying
            from a source list.

    Stability:
        Caller
    """

    ordered: bool = False
    "Whether this list has a meaningful order."

    _len: int
    "Length of the item list."

    # storage for individual components of the item list
    _ids: pa.Array | None = None
    _ids_numpy: IDArray | None = None
    _mask_cache: tuple[ItemList, np.ndarray[tuple[int], np.dtype[np.bool_]]] | None = None
    _numbers: MTArray[np.int32] | None = None
    _vocab: Vocabulary | None = None
    _ranks: MTArray[np.int32] | None = None
    _fields: dict[str, MTGenericArray]

    @overload
    def __init__(
        self,
        source: ItemList,
        *,
        ordered: bool | None = None,
        vocabulary: Vocabulary | None = None,
        scores: NDArray[np.generic]
        | torch.Tensor
        | ArrayLike
        | Literal[False]
        | np.floating
        | float
        | None = None,
        **fields: NDArray[np.generic] | torch.Tensor | ArrayLike | Literal[False],
    ): ...
    @overload
    def __init__(
        self,
        source: IDSequence | None = None,
        *,
        item_nums: NDArray[np.int32] | pd.Series[int] | Sequence[int] | ArrayLike | None = None,
        vocabulary: Vocabulary | None = None,
        ordered: bool | None = None,
        scores: NDArray[np.generic]
        | torch.Tensor
        | ArrayLike
        | Literal[False]
        | np.floating
        | float
        | None = None,
        **fields: NDArray[np.generic] | torch.Tensor | ArrayLike | Literal[False],
    ): ...
    @overload
    def __init__(
        self,
        source: None = None,
        *,
        item_ids: IDSequence | None = None,
        item_nums: NDArray[np.int32] | pd.Series[int] | Sequence[int] | ArrayLike | None = None,
        vocabulary: Vocabulary | None = None,
        ordered: bool | None = None,
        scores: NDArray[np.generic]
        | torch.Tensor
        | ArrayLike
        | Literal[False]
        | np.floating
        | float
        | None = None,
        **fields: NDArray[np.generic] | torch.Tensor | ArrayLike | Literal[False],
    ): ...
    def __init__(  # type: ignore
        self,
        source: ItemList | IDSequence | None = None,
        *,
        ordered: bool | None = None,
        vocabulary: Vocabulary | None = None,
        _init_dict: dict[str, Any] | None = None,
        _init_array: pa.StructArray | None = None,
        **kwargs: Any,
    ):
        if _init_dict is not None:
            self.__dict__.update(_init_dict)
            return
        elif _init_array is not None:
            self._init_from_arrow(_init_array, vocabulary, ordered)
        elif isinstance(source, ItemList):
            self.__dict__.update(source.__dict__)
            if ordered is not None:
                self.ordered = ordered
            elif "rank" in kwargs:
                self.ordered = True

            if vocabulary is not None:  # pragma: nocover
                raise ValueError("cannot change vocabulary from item list")

            self._init_fields(**kwargs)
        else:
            self.ordered = ordered if ordered is not None else ("rank" in kwargs)
            self._vocab = vocabulary
            self._init_ids(source, **kwargs)
            self._init_numbers(**kwargs)
            self._init_check_length()
            self._init_fields(**kwargs)

        if vocabulary is not None:
            if not isinstance(vocabulary, Vocabulary):  # pragma: nocover
                raise TypeError(f"expected Vocabulary, got {type(vocabulary)}")
            self._vocab = vocabulary

    def _init_from_arrow(
        self, array: pa.StructArray, vocabulary: Vocabulary | None, ordered: bool | None
    ):
        self._len = len(array)
        self._vocab = vocabulary

        # get the ID and number fields
        try:
            self._ids = array.field("item_id")
        except KeyError:
            pass

        try:
            numbers = array.field("item_num")
        except KeyError:
            pass
        else:
            self._numbers = MTArray(numbers.cast(pa.int32()))

        # set up ranking and ordering
        try:
            ranks = array.field("rank")
        except KeyError:
            self.ordered = ordered or False
        else:
            if ordered is False:  # pragma: nocover
                raise ValueError("table has ranks but ordered=False")
            self.ordered = True
            self._ranks = MTArray(ranks)

        # the rest of the fields can be lazily loaded from the array
        fields = [array.type.field(i) for i in range(array.type.num_fields)]
        self._fields = {
            f.name: MTArray(array.field(i))
            for (i, f) in enumerate(fields)
            if f.name not in ["item_num", "item_id"]
        }

    def _init_ids(
        self,
        source: IDSequence | None,
        *,
        item_ids=None,
        item_id=None,
        **ignore,
    ):
        """
        Initialize the item list's item IDs and length.  The vocabulary should
        already be set.
        """
        # handle aliases for item ID columns
        if source is not None:
            if item_ids is not None or item_id is not None:
                raise ValueError("cannot specify both item_ids & item ID source")
            item_ids = source
        elif item_id is not None:
            if item_ids is not None:
                raise ValueError("cannot specify both item_ids & item_id")
            item_ids = item_id
        elif item_ids is None:
            # no setup
            return

        # handle the item ID type
        if isinstance(item_ids, pa.Array):
            self._ids = item_ids
        elif len(item_ids) > 0:
            try:
                self._ids = pa.array(item_ids)  # type: ignore
            except pa.ArrowInvalid:  # pragma: nocover
                raise TypeError("invalid item ID type or dimension")
            if isinstance(item_ids, np.ndarray):
                self._ids_numpy = item_ids
        else:
            return

        assert self._ids is not None
        idt = self._ids.type
        if not (
            pa.types.is_integer(idt)
            or pa.types.is_string(idt)
            or pa.types.is_binary(idt)
            or pa.types.is_large_string(idt)
            or pa.types.is_large_binary(idt)
        ):
            raise TypeError(f"item IDs not integers or strings (type: {idt})")

        self._len = len(item_ids)

    def _init_numbers(self, item_nums=None, item_num=None, **ignore):
        """
        Initialize the item numbers.
        """

        if item_num is not None:
            if item_nums is not None:
                raise ValueError("cannot specify both item_nums and item_num")
            item_nums = item_num

        length = getattr(self, "_len", None)

        if item_nums is None:
            return

        if not len(item_nums):  # type: ignore
            item_nums = np.ndarray(0, dtype=np.int32)
        if isinstance(item_nums, np.ndarray):
            nk = item_nums.dtype.kind
            if nk != "i" and nk != "u":
                raise TypeError(f"invalid number dtype {item_nums.dtype}")
            item_nums = np.require(item_nums, dtype=np.int32)
        elif isinstance(item_nums, pa.Array):
            if not pa.types.is_integer(item_nums.type):
                raise TypeError(f"invalid number type {item_nums.type}")
            item_nums = item_nums.cast(pa.int32())
        elif torch.is_tensor(item_nums):
            item_nums = item_nums.to(torch.int32)

        self._numbers = MTArray(item_nums)
        check_1d(self._numbers, length, label="item_nums")
        self._len = self._numbers.shape[0]

    def _init_check_length(self):
        """
        Check that we have a correct length set.

        This is called after setting IDs and numbers, and before setting fields.
        After this succeeds, ``_len`` is set and ``len(self)`` will work.
        """
        length = getattr(self, "_len", None)
        if self._ids is None and self._numbers is None:
            self._ids = pa.array([], pa.int32())
            self._numbers = MTArray(pa.array([], pa.int32()))
            self._len = 0
        elif length is None:  # pragma: nocover
            # this is here as a fail-safe, but isn't triggered in realistic scenarios
            if self._ids is not None:
                self._len = len(self._ids)
            elif self._numbers is not None:
                self._len = self._numbers.shape[0]
            else:
                raise AssertionError("unreachable")

    def _init_fields(self, *, scores=None, score=None, rank=None, **other):
        if any(k[0] == "_" for k in other.keys()):  # pragma: nocover
            raise ValueError("item list fields cannot start with _")

        fields = getattr(self, "_fields", {}).copy()

        if score is not None:
            if scores is not None:
                raise ValueError("cannot specify both score= and scores=")
            scores = score

        if scores is False and "score" in fields:  # check 'is False' to distinguish from None
            del fields["score"]
        elif scores is not None:
            if np.isscalar(scores):
                scores = np.full(self._len, scores, dtype=np.float32)
            elif isinstance(scores, np.ndarray):
                scores = np.require(scores, np.float32)
            elif isinstance(scores, torch.Tensor):
                scores = scores.to(torch.float32)

            fields["score"] = check_1d(MTArray(scores), self._len, label="scores")

        if rank is not None:
            if not self.ordered:
                warnings.warn(
                    "ranks provided but ordered=False, dropping ranks", DataWarning, stacklevel=3
                )
            else:
                self._ranks = check_1d(
                    MTArray(np.require(rank, np.int32)), self._len, label="ranks"
                )
                fields["rank"] = self._ranks
                if self._len and self._ranks.numpy()[0] != 1:
                    warnings.warn("ranks do not begin with 1", DataWarning, stacklevel=3)

        # convert remaining fields
        for name, fdata in other.items():
            if name in ("item_id", "item_num", "item_ids", "item_nums", "score", "rank"):
                continue
            if fdata is False:
                if name in fields:
                    del fields[name]
                continue

            if fdata is not None:
                fields[name] = check_1d(MTArray(fdata), self._len, label=name)

        self._fields = fields

    @classmethod
    def from_df(
        cls, df: pd.DataFrame, *, vocabulary: Vocabulary | None = None, keep_user: bool = False
    ) -> ItemList:
        """
        Create a item list from a Pandas data frame.  The frame should have
        ``item_num`` and/or ``item_id`` columns to identify the items; other
        columns (e.g. ``score`` or ``rating``) are added as fields. If the data
        frame has user columns (``user_id`` or ``user_num``), those are dropped
        by default.

        Args:
            df:
                The data frame to turn into an item list.
            vocabulary:
                The item vocabulary.
            keep_user:
                If ``True``, keeps user ID/number columns instead of dropping them.
        """
        ids = df["item_id"].values if "item_id" in df.columns else None
        nums = df["item_num"].values if "item_num" in df.columns else None
        if ids is None and nums is None:
            raise TypeError("data frame must have at least one of item_id, item_num columns")

        to_drop = ["item_id", "item_num"]
        if not keep_user:
            to_drop += ["user_id", "user_num"]

        fields = {f: df[f].values for f in df.columns if f not in to_drop}

        items = cls(
            item_ids=ids,  # type: ignore
            item_nums=nums,  # type: ignore
            vocabulary=vocabulary,
            **fields,  # type: ignore
        )

        return items

    @classmethod
    def from_arrow(
        cls,
        tbl: pa.StructArray | pa.ChunkedArray | pa.Table,
        *,
        vocabulary: Vocabulary | None = None,
    ) -> ItemList:
        """
        Create a item list from a Pandas table or structured array.  The table
        should have ``item_num`` and/or ``item_id`` columns to identify the
        items; other columns (e.g. ``score`` or ``rating``) are added as fields.
        If the data frame has user columns (``user_id`` or ``user_num``), those
        are dropped by default.

        Args:
            tbl:
                The Arrow table or array to convert to an item list.
            vocabulary:
                The item vocabulary.
        """
        if isinstance(tbl, pa.Table):
            if tbl.num_rows:
                tbl = tbl.to_struct_array()
            else:
                tbl = pa.array([], pa.struct(tbl.schema))

        if isinstance(tbl, pa.ChunkedArray):
            tbl = tbl.combine_chunks()  # type: ignore
        assert isinstance(tbl, pa.StructArray)
        assert isinstance(tbl.type, pa.StructType)

        if hasattr(tbl.type, "names"):
            names = tbl.type.names  # type: ignore
        else:
            names = [tbl.type.field(i).name for i in range(tbl.type.num_fields)]

        if "item_id" not in names and "item_num" not in names:
            raise TypeError("data table must have at least one of item_id, item_num columns")

        return ItemList(_init_array=tbl, vocabulary=vocabulary)

    @classmethod
    def from_vocabulary(cls, vocab: Vocabulary) -> ItemList:
        return ItemList(
            item_ids=vocab.id_array(), item_nums=np.arange(len(vocab)), vocabulary=vocab
        )

    def clone(self) -> ItemList:
        """
        Make a shallow copy of the item list.
        """
        return ItemList(
            _init_dict={
                "_len": self._len,
                "_ids": self._ids,
                "_numbers": self._numbers,
                "_vocab": self._vocab,
                "ordered": self.ordered,
                "_ranks": self._ranks,
                "_fields": self._fields,
            }  # type: ignore
        )  # type: ignore

    @property
    def vocabulary(self) -> Vocabulary | None:
        "Get the item list's vocabulary, if available."
        return self._vocab

    @overload
    def ids(self, *, format: Literal["numpy"] = "numpy") -> IDArray: ...
    @overload
    def ids(self, *, format: Literal["arrow"]) -> pa.Array: ...
    def ids(self, *, format: Literal["numpy", "arrow"] = "numpy"):
        """
        Get the item IDs.

        Returns:
            An array of item identifiers.

        Raises:
            RuntimeError: if the item list was not created with IDs or a :class:`Vocabulary`.
        """
        if self._ids is None:
            if self._vocab is None:
                raise RuntimeError("item IDs not available (no IDs or vocabulary provided)")
            assert self._numbers is not None
            self._ids_numpy = self._vocab.ids(self._numbers.numpy())
            self._ids = pa.array(self._ids_numpy)

        match format:
            case "numpy":
                if self._ids_numpy is None:
                    self._ids_numpy = self._ids.to_numpy(zero_copy_only=False)
                return self._ids_numpy
            case "arrow":
                return self._ids
            case _:  # pragma: nocover
                raise ValueError(f"unknown format {format}")

    @overload
    def numbers(
        self,
        format: Literal["numpy"] = "numpy",
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> NPVector[np.int32]: ...
    @overload
    def numbers(
        self,
        format: Literal["torch"],
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> torch.Tensor: ...
    @overload
    def numbers(
        self,
        format: Literal["arrow"],
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> pa.Array[pa.Int32Scalar]: ...
    @overload
    def numbers(
        self,
        format: LiteralString = "numpy",
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> ArrayLike: ...
    def numbers(
        self,
        format: LiteralString = "numpy",
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> ArrayLike:
        """
        Get the item numbers.

        Args:
            format:
                The array format to use.
            vocabulary:
                A alternate vocabulary for mapping IDs to numbers.  If provided,
                then the item list must have IDs (either stored, or through a
                vocabulary).

        Returns:
            An array of item numbers.

        Raises:
            RuntimeError: if the item list was not created with numbers or a
            :class:`Vocabulary`.
        """
        if vocabulary is not None and vocabulary != self._vocab:
            # we need to translate vocabulary
            ids = self.ids()
            mta = MTArray(vocabulary.numbers(ids, missing=missing))
            return mta.to(format)

        if self._numbers is None:
            if self._vocab is None:
                raise RuntimeError("item numbers not available (no IDs or vocabulary provided)")
            assert self._ids is not None
            self._numbers = MTArray(self._vocab.numbers(self._ids, missing="negative"))

        if missing == "error" and np.any(self._numbers.numpy() < 0):
            raise KeyError("item IDs")
        return self._numbers.to(format)

    @overload
    def scores(self, format: Literal["numpy"] = "numpy") -> NPVector | None: ...
    @overload
    def scores(self, format: Literal["torch"]) -> torch.Tensor | None: ...
    @overload
    def scores(self, format: Literal["arrow"]) -> pa.Array[pa.FloatScalar] | None: ...
    @overload
    def scores(
        self, format: Literal["pandas"], *, index: Literal["ids", "numbers"] | None = None
    ) -> pd.Series | None: ...
    @overload
    def scores(self, format: LiteralString = "numpy") -> ArrayLike | None: ...
    def scores(self, format: LiteralString = "numpy", **kwargs) -> ArrayLike | None:
        """
        Get the item scores (if available).
        """
        return self.field("score", format=format, **kwargs)

    @overload
    def ranks(self, format: Literal["numpy"] = "numpy") -> NDArray[np.int32] | None: ...
    @overload
    def ranks(self, format: Literal["torch"]) -> torch.Tensor | None: ...
    @overload
    def ranks(self, format: Literal["arrow"]) -> pa.Array[pa.Int32Scalar] | None: ...
    @overload
    def ranks(self, format: Literal["pandas"]) -> pd.Series[int] | None: ...
    @overload
    def ranks(self, format: LiteralString = "numpy") -> ArrayLike | None: ...
    def ranks(self, format: LiteralString = "numpy") -> ArrayLike | None:
        """
        Get an array of ranks for the items in this list, if it is ordered.
        Unordered lists have no ranks.  The ranks are based on the order in the
        list, **not** on the score.

        Item ranks start with **1**, for compatibility with common practice in
        mathematically defining information retrieval metrics and operations.

        Returns:
            An array of item ranks, or ``None`` if the list is unordered.
        """
        if not self.ordered:
            return None

        if self._ranks is None:
            self._ranks = MTArray(np.arange(1, self._len + 1, dtype=np.int32))

        if format == "pandas":
            ranks = self._ranks.to("numpy")
            return pd.Series(ranks, index=self.ids())
        else:
            return self._ranks.to(format)

    @overload
    def field(
        self, name: str, format: Literal["numpy"] = "numpy"
    ) -> NDArray[np.floating] | None: ...
    @overload
    def field(self, name: str, format: Literal["torch"]) -> torch.Tensor | None: ...
    @overload
    def field(self, name: str, format: Literal["arrow"]) -> pa.Array | pa.Tensor | None: ...
    @overload
    def field(
        self,
        name: str,
        format: Literal["pandas"],
        *,
        index: Literal["ids", "numbers"] | None = None,
    ) -> pd.Series | None: ...
    @overload
    def field(self, name: str, format: Literal["multi"]) -> MTArray | None: ...
    @overload
    def field(self, name: str, format: LiteralString) -> ArrayLike | None: ...
    def field(
        self, name: str, format: LiteralString = "numpy", *, index: LiteralString | None = None
    ) -> object:
        val = self._fields.get(name, None)

        if val is None:
            return None
        elif format == "multi":
            return val
        elif format == "pandas":
            idx = None
            vs = val.to("numpy")
            if index == "ids":
                idx = pd.Index(self.ids(), name="item_id")
            elif index == "numbers":
                idx = pd.Index(self.numbers(), name="item_num")
            elif index:  # pragma: nocover
                raise ValueError(f"unsupported Pandas index {index}")
            return pd.Series(vs, index=idx)
        else:
            return val.to(format)

    def isin(self, other: ItemList) -> np.ndarray[tuple[int], np.dtype[np.bool_]]:
        """
        Return a boolean mask identifying the items of this list that are in the
        other list.

        This is equivalent to :func:`numpy.isin` applied to the ID arrays, but
        is much more efficient in many cases.
        """
        # cache is-in checks for identical arrays
        if self._mask_cache is not None and self._mask_cache[0] is other:
            return self._mask_cache[1]

        # fast path — try to just use a mask.
        if self.vocabulary is not None:
            # In most common hot-path cases, left is at least as long as right.
            mask = np.zeros(len(self.vocabulary), dtype=np.bool_)
            onums = other.numbers(vocabulary=self.vocabulary, missing="negative")
            onums = onums[onums >= 0]
            mask[onums] = True
            nums = self.numbers()
            result = mask[nums]
        else:
            id_arr = self.ids(format="arrow")
            if pa.types.is_integer(id_arr.type):
                # numpy is quicker than arrow for integer comparisons
                result = np.isin(id_arr.to_numpy(), other.ids())
            else:
                # arrow is quicker than numpy for strings
                try:
                    result = pc.is_in(self.ids(format="arrow"), other.ids(format="arrow")).to_numpy(
                        zero_copy_only=False
                    )
                except pa.ArrowTypeError:
                    # ids are of incompatible types, nothing matches
                    result = np.zeros(len(self), dtype=np.bool_)

        self._mask_cache = (other, result)
        return result

    def to_df(self, *, ids: bool = True, numbers: bool = True) -> pd.DataFrame:
        """
        Convert this item list to a Pandas data frame.  It has the following columns:

        * ``item_id`` — the item IDs (if available and ``ids=True``)
        * ``item_num`` — the item numbers (if available and ``numbers=True``)
        * ``score`` — the item scores
        * ``rank`` — the item ranks (if the list is ordered)
        * all other defined fields, using their field names
        """
        cols = {}
        if ids and (self._ids is not None or self._vocab is not None):
            cols["item_id"] = self.ids()
        if numbers and (self._numbers is not None or self._vocab is not None):
            cols["item_num"] = self.numbers()
        # we need to have numbers or ids, or it makes no sense
        if "item_id" not in cols and "item_num" not in cols:
            if ids and not numbers:
                raise RuntimeError("item list has no vocabulary, cannot compute IDs")
            elif numbers and not ids:
                raise RuntimeError("item list has no vocabulary, cannot compute numbers")
            else:
                raise RuntimeError("cannot create item data frame without identifiers or numbers")

        score = self.scores()
        if score is not None:
            cols["score"] = score
        if self.ordered:
            cols["rank"] = self.ranks()

        # add remaining fields
        cols.update((k, v.numpy()) for (k, v) in self._fields.items() if k not in ("score", "rank"))
        return pd.DataFrame(cols)

    @overload
    def to_arrow(
        self,
        *,
        ids: bool = True,
        numbers: bool = False,
        type: Literal["table"] = "table",
        columns: dict[str, pa.DataType] | None = None,
    ) -> pa.Table: ...
    @overload
    def to_arrow(
        self,
        *,
        ids: bool = True,
        numbers: bool = False,
        type: Literal["array"],
        columns: dict[str, pa.DataType] | None = None,
    ) -> pa.StructArray: ...
    def to_arrow(
        self,
        *,
        ids: bool = True,
        numbers: bool = False,
        type: Literal["table", "array"] = "table",
        columns: dict[str, pa.DataType] | None = None,
    ):
        """
        Convert the item list to a Pandas table.
        """
        arrays = []
        names = []
        if columns is not None and len(self) == 0:
            arrays = [pa.array([], ft) for ft in columns.values()]
            names = list(columns.keys())
        else:
            if columns is None:
                columns = self.arrow_types(ids=ids, numbers=numbers)

            for c_name, c_type in columns.items():
                names.append(c_name)
                if c_name == "item_id":
                    arrays.append(self.ids(format="arrow"))
                elif c_name == "item_num":
                    arrays.append(self.numbers("arrow"))
                elif c_name == "rank":
                    if self.ordered:
                        arrays.append(self.ranks("arrow"))
                    else:
                        warnings.warn("requested rank column for unordered list", DataWarning)
                        arrays.append(pa.nulls(len(self), c_type))
                else:
                    fld = self.field(c_name, format="arrow")

                    if fld is not None:
                        arrays.append(fld)
                    else:
                        warnings.warn(f"unknown field {c_name}", DataWarning)
                        arrays.append(pa.nulls(len(self), c_type))

        if type == "table":
            return pa.Table.from_arrays(arrays, names)
        elif type == "array":
            return pa.StructArray.from_arrays(arrays, names)
        else:  # pragma: nocover
            raise ValueError(f"unsupported target type {type}")

    def arrow_types(self, *, ids: bool = True, numbers: bool = False) -> dict[str, pa.DataType]:
        """
        Get the Arrow data types for this item list.
        """
        types: dict[str, pa.DataType] = {}

        if ids:
            if self._ids is not None:
                types["item_id"] = self._ids.type
            elif self._vocab is not None:
                types["item_id"] = self._vocab.id_array().type

        if numbers and (self._numbers is not None or self._vocab is not None):
            types["item_num"] = pa.int32()

        if self.ordered:
            types["rank"] = pa.int32()

        for n, f in self._fields.items():
            types[n] = f.arrow().type

        return types

    def top_n(self, n: int | None = None, *, scores: str | NPVector | None = None) -> ItemList:
        """
        Get the top _N_ items in this list, sorted in decreasing order.

        If any scores are undefined (``NaN``), those items are excluded.

        Args:
            n:
                The number of items.  If ``None`` or negative, returns all items
                sorted by score.
            scores:
                The name of a field containing the scores, or a NumPy vector of
                scores, for selecting the top _N_.  If ``None``, the item list's
                scores are used.
        Returns:
            An ordered item list containing the top ``n`` items.
        """
        score_tensor = None
        if scores is None:
            scores = self.field("score", format="multi")
            if scores is None:
                raise ValueError("item list has no scores")
            score_tensor = scores.torch(convert=False)
        elif isinstance(scores, str):
            scores = self.field(scores, format="multi")
            if scores is None:
                raise KeyError(scores)
            score_tensor = scores.torch(convert=False)
        elif len(scores) != len(self):
            raise ValueError("score array must have same length as items")
        elif isinstance(scores, torch.Tensor):
            score_tensor = scores

        want_all = n is None or n < 0

        if want_all and score_tensor is not None and not score_tensor.is_cpu:
            # sorting is faster on-device
            vs, picked = torch.sort(score_tensor, descending=True)

            # drop invalid / unscored items
            # on large arrays, actually faster to do the final filter in NumPy.
            vs = vs.cpu().numpy()
            picked = picked.cpu().numpy()
            invalid = np.isnan(vs)
            picked = picked[~invalid]
        else:
            if not isinstance(scores, MTArray):
                scores = MTArray(scores)
            scores = scores.arrow()

            if want_all:
                picked = _data_accel.argsort_descending(scores)
            else:
                picked = argtopn(scores, n)

        return self._take(picked, ordered=True)

    @overload
    def remove(
        self,
        *,
        ids: IDSequence | None = None,
    ) -> ItemList: ...
    @overload
    def remove(
        self,
        *,
        numbers: pa.Int32Array | NPVector[np.integer] | pd.Series[int] | None = None,
    ) -> ItemList: ...
    def remove(
        self,
        *,
        ids: IDSequence | None = None,
        numbers: pa.Int32Array | NPVector[np.integer] | pd.Series[int] | None = None,
    ) -> ItemList:
        """
        Return an item list with the specified items removed.

        The items to remove are not required to be in the list.

        Args:
            ids:
                The item IDs to remove.
            numbers:
                The item numbers to remove.
        """
        if ids is None and numbers is None:
            raise ValueError("must specify one of ids= or numbers=")

        mask = None
        if ids is not None:
            if numbers is not None:
                raise ValueError("must specify only one of ids= or numbers=")

            if self._vocab is not None:
                numbers = self._vocab.numbers(ids, missing="negative")
                numbers = numbers[numbers >= 0]
            else:
                # handle IDs the slow way
                mask = ~self.isin(ItemList(item_ids=ids))

        if mask is None:
            assert numbers is not None

            numbers = MTArray(numbers).numpy()
            numbers = np.require(numbers, dtype=np.int32)
            mask = np.isin(self.numbers(), numbers, invert=True, kind="table")

        # fast case — we have a vocabulary and no fields
        if self._vocab is not None and not self._fields:
            return ItemList(
                item_nums=self.numbers()[mask], vocabulary=self.vocabulary, ordered=self.ordered
            )

        return self._take(mask)

    def _take(self, sel: ILIndexer, *, ordered: bool | None = None) -> ItemList:
        """
        Implementation helper for indexing.
        """
        indexer = get_indexer(sel)

        if ordered is None:
            ordered = self.ordered

        # Only subset the IDs if we don't have a vocabulary.  Otherwise, defer
        # ID subset until IDs are actually needed.
        array = self.to_arrow(ids=self.vocabulary is None, numbers=True, type="array")
        array = indexer(array)

        return ItemList(_init_array=array, vocabulary=self.vocabulary, ordered=ordered)  # type: ignore

    def __len__(self):
        return self._len

    def __getitem__(self, sel: ILIndexer) -> ItemList:
        """
        Subset the item list.

        .. todo::
            Support on-device masking.

        Args:
            sel:
                The items to select. Can be either a Boolean array of the same
                length as the list that is ``True`` to indicate selected items,
                or an array of indices of the items to retain (in order in the
                list, starting from 0).
        """
        return self._take(sel)

    def __getstate__(self) -> dict[str, object]:
        state: dict[str, object] = {"ordered": self.ordered, "len": self._len}
        if self._ids is not None:
            state["ids"] = self._ids
        elif self._vocab is not None:
            # compute the IDs so we can save them
            state["ids"] = self.ids(format="arrow")

        if self._numbers is not None:
            state["numbers"] = self._numbers.numpy()
        elif self._vocab is not None:
            state["numbers"] = self.numbers()

        state.update(("field_" + k, v.numpy()) for (k, v) in self._fields.items())
        return state

    def __setstate__(self, state: dict[str, Any]):
        self.ordered = state["ordered"]
        self._len = state["len"]
        self._ids = state.get("ids", None)
        if "numbers" in state:
            self._numbers = MTArray(state["numbers"])
        self._fields = {k[6:]: MTArray(v) for (k, v) in state.items() if k.startswith("field_")}

    def __str__(self) -> str:  # pragma: nocover
        return f"<ItemList of {self._len} items>"

    def __repr__(self) -> str:  # pragma: nocover
        out = io.StringIO()
        fields = self._fields
        if self.ordered and "rank" not in fields:
            fields = fields | {"rank": None}
        nf = len(fields)
        print(f"<ItemList of {self._len} items with {nf} fields", "{", file=out)

        if self._ids is not None:
            print("  ids:", np.array2string(self.ids(), threshold=10), file=out)
        elif self._vocab is not None:
            print("  ids: <lazy>", file=out)
        if self._numbers is not None:
            print("  numbers:", np.array2string(self._numbers.numpy(), threshold=10), file=out)
        elif self._vocab is not None:
            print("  numbers: <lazy>", file=out)
        for name, f in sorted(fields.items(), key=lambda t: t[0]):
            if f is None:
                print(f"  {name}: <lazy>", file=out)
            else:
                print(f"  {name}:", np.array2string(f.numpy(), threshold=10), file=out)
        print("}>", end="", file=out)

        return out.getvalue()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pytest import raises

from lenskit.pipeline import PipelineBuilder
from lenskit.pipeline.types import Lazy


def proc_hello(msg: str) -> str:
    return f"Hello, {msg}"


def proc_incr(x: int) -> int:
    return x + 1


def proc_exclaim(msg: str) -> str:
    return msg + "!"


def lazy_exclaim(msg: Lazy[str], other: Lazy[int]) -> str:
    return msg.get() + "!"


def test_raise_invalid_input():
    build = PipelineBuilder()
    msg = build.create_input("message", str)
    hello = build.add_component("hello", proc_hello, msg=msg)

    pipe = build.build()

    out = pipe.run(hello, message="world")
    assert out == "Hello, world"

    with raises(TypeError, match="invalid data for input"):
        pipe.run(hello, message=5)


def test_raise_invalid_wired_input():
    build = PipelineBuilder()
    x = build.create_input("x", int)
    incr = build.add_component("incr", proc_incr, x=x)
    excl = build.add_component("exclaim", proc_exclaim, msg=incr)

    pipe = build.build()

    with raises(TypeError, match="found.*, expected"):
        pipe.run(excl, x=5)


def test_raise_lazy_input():
    build = PipelineBuilder()
    x = build.create_input("x", int)
    incr = build.add_component("incr", proc_incr, x=x)
    excl = build.add_component("exclaim", lazy_exclaim, msg=incr, other="bob")

    pipe = build.build()

    with raises(TypeError, match="found.*, expected"):
        pipe.run(excl, x=5)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import approx, raises, warns

from lenskit.data import DatasetBuilder
from lenskit.data.dataset import Dataset
from lenskit.data.schema import AllowableTroolean, DataSchema
from lenskit.diagnostics import DataError, DataWarning


def test_save_load_ml(ml_ratings: pd.DataFrame, tmpdir: Path):
    dsb = DatasetBuilder()
    dsb.add_interactions("rating", ml_ratings, entities=["user", "item"], missing="insert")

    ds = dsb.build()

    ds_path = tmpdir / "dataset"
    dsb.save(ds_path)

    assert ds_path.exists()

    schema_path = ds_path / "schema.json"
    assert schema_path.exists()
    schema = DataSchema.model_validate_json(schema_path.read_text("utf8"))
    assert schema == dsb.schema

    assert (ds_path / "item.parquet").exists()
    assert (ds_path / "user.parquet").exists()
    assert (ds_path / "rating.parquet").exists()

    ds2 = Dataset.load(ds_path)

    assert ds2.user_count == ds.user_count
    assert ds2.item_count == ds.item_count
    assert np.all(ds2.users.ids() == ds.users.ids())
    assert np.all(ds2.items.ids() == ds.items.ids())
    assert ds2.interaction_count == ds.interaction_count



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from itertools import product
from typing import Literal

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import approx, mark, raises, warns

from lenskit.data import Dataset, DatasetBuilder
from lenskit.data.schema import AllowableTroolean
from lenskit.diagnostics import DataError, DataWarning


@dataclass
class QueryDT:
    s: str
    dt: datetime
    ts: float
    thresh: str | datetime | float
    compare: datetime | float

    @classmethod
    def create(
        cls,
        spec: str,
        t_fmt: Literal["datetime", "timestamp", "string"],
        d_fmt: Literal["int", "timestamp"],
    ) -> QueryDT:
        dt = datetime.fromisoformat(spec)
        ts = dt.timestamp()
        match t_fmt:
            case "datetime":
                thresh = dt
            case "timestamp":
                thresh = ts
            case "string":
                thresh = spec

        match d_fmt:
            case "int":
                comp = ts
            case "timestamp":
                comp = dt

        return cls(s=spec, dt=dt, ts=ts, thresh=thresh, compare=comp)


@mark.parametrize(
    ["ts_fmt", "q_fmt"], product(["int", "timestamp"], ["datetime", "timestamp", "string"])
)
def test_filter_ratings_min_time(
    ts_fmt: Literal["int", "timestamp"],
    q_fmt: Literal["datetime", "timestamp", "string"],
    ml_ratings: pd.DataFrame,
):
    dsb = DatasetBuilder()
    if ts_fmt == "int":
        ml_ratings = ml_ratings.assign(
            timestamp=(ml_ratings["timestamp"] - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")
        )
    dsb.add_interactions(
        "rating", ml_ratings, entities=["user", "item"], missing="insert", default=True
    )
    q = QueryDT.create("2001-01-01", q_fmt, ts_fmt)
    dsb.filter_interactions("rating", min_time=q.thresh)
    ds = dsb.build()
    assert ds.interactions().pandas()["timestamp"].min() >= q.compare
    assert ds.interactions().pandas()["timestamp"].max() == ml_ratings["timestamp"].max()


@mark.parametrize(
    ["ts_fmt", "q_fmt"], product(["int", "timestamp"], ["datetime", "timestamp", "string"])
)
def test_filter_ratings_max_time(
    ts_fmt: Literal["int", "timestamp"],
    q_fmt: Literal["datetime", "timestamp", "string"],
    ml_ratings: pd.DataFrame,
):
    dsb = DatasetBuilder()
    if ts_fmt == "int":
        ml_ratings = ml_ratings.assign(
            timestamp=(ml_ratings["timestamp"] - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")
        )
    dsb.add_interactions(
        "rating", ml_ratings, entities=["user", "item"], missing="insert", default=True
    )
    q = QueryDT.create("2001-01-01", q_fmt, ts_fmt)
    dsb.filter_interactions("rating", max_time=q.thresh)
    ds = dsb.build()
    assert ds.interactions().pandas()["timestamp"].min() == ml_ratings["timestamp"].min()
    assert ds.interactions().pandas()["timestamp"].max() < q.compare


@mark.parametrize(
    ["ts_fmt", "q_fmt"], product(["int", "timestamp"], ["datetime", "timestamp", "string"])
)
def test_filter_ratings_min_max_time(
    ts_fmt: Literal["int", "timestamp"],
    q_fmt: Literal["datetime", "timestamp", "string"],
    ml_ratings: pd.DataFrame,
):
    dsb = DatasetBuilder()
    if ts_fmt == "int":
        ml_ratings = ml_ratings.assign(
            timestamp=(ml_ratings["timestamp"] - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")
        )
    dsb.add_interactions(
        "rating", ml_ratings, entities=["user", "item"], missing="insert", default=True
    )
    q1 = QueryDT.create("2001-01-01", q_fmt, ts_fmt)
    q2 = QueryDT.create("2004-01-01", q_fmt, ts_fmt)
    dsb.filter_interactions("rating", min_time=q1.thresh, max_time=q2.thresh)
    ds = dsb.build()
    assert ds.interactions().pandas()["timestamp"].min() >= q1.compare
    assert ds.interactions().pandas()["timestamp"].max() < q2.compare


def test_filter_ratings_nums_df(rng, ml_ds: Dataset):
    dsb = DatasetBuilder(ml_ds)

    rates = ml_ds.interactions().pandas(attributes=[])
    samp = rates.sample(1000, replace=False, random_state=rng, ignore_index=True)

    dsb.filter_interactions("rating", remove=samp)

    ds = dsb.build()
    assert ds.interaction_count == ml_ds.interaction_count - 1000

    orig_istats = ml_ds.item_stats()
    new_istats = ds.item_stats()
    si_counts = samp["item_num"].value_counts()
    si_counts = pd.Series(si_counts.values, index=ml_ds.items.ids(si_counts.index.values))
    si_counts = si_counts.reindex(new_istats.index, fill_value=0)

    assert np.all(new_istats["count"] == orig_istats["count"] - si_counts)


def test_filter_ratings_nums_tbl(rng, ml_ds: Dataset):
    dsb = DatasetBuilder(ml_ds)

    rates = ml_ds.interactions().pandas(attributes=[])
    samp_df = rates.sample(1000, replace=False, random_state=rng, ignore_index=True)
    samp = pa.Table.from_pandas(samp_df, preserve_index=False)

    dsb.filter_interactions("rating", remove=samp)

    ds = dsb.build()
    assert ds.interaction_count == ml_ds.interaction_count - 1000

    orig_istats = ml_ds.item_stats()
    new_istats = ds.item_stats()
    si_counts = samp_df["item_num"].value_counts()
    si_counts = pd.Series(si_counts.values, index=ml_ds.items.ids(si_counts.index.values))
    si_counts = si_counts.reindex(new_istats.index, fill_value=0)

    assert np.all(new_istats["count"] == orig_istats["count"] - si_counts)


def test_filter_ratings_ids_df(rng, ml_ds: Dataset):
    dsb = DatasetBuilder(ml_ds)

    rates = ml_ds.interactions().pandas(attributes=[], ids=True)
    samp = rates.sample(1000, replace=False, random_state=rng, ignore_index=True)

    dsb.filter_interactions("rating", remove=samp)

    ds = dsb.build()
    assert ds.interaction_count == ml_ds.interaction_count - 1000

    orig_istats = ml_ds.item_stats()
    new_istats = ds.item_stats()
    si_counts = samp["item_id"].value_counts()
    si_counts = si_counts.reindex(new_istats.index, fill_value=0)

    assert np.all(new_istats["count"] == orig_istats["count"] - si_counts)


def test_filter_ratings_ids_tbl(rng, ml_ds: Dataset):
    dsb = DatasetBuilder(ml_ds)

    rates = ml_ds.interactions().pandas(attributes=[], ids=True)
    samp_df = rates.sample(1000, replace=False, random_state=rng, ignore_index=True)
    samp = pa.Table.from_pandas(samp_df, preserve_index=False)

    dsb.filter_interactions("rating", remove=samp)

    ds = dsb.build()
    assert ds.interaction_count == ml_ds.interaction_count - 1000

    orig_istats = ml_ds.item_stats()
    new_istats = ds.item_stats()
    si_counts = samp_df["item_id"].value_counts()
    si_counts = si_counts.reindex(new_istats.index, fill_value=0)

    assert np.all(new_istats["count"] == orig_istats["count"] - si_counts)


def test_filter_ratings_readd(rng, ml_ds: Dataset):
    "check that adding again after removal works"
    dsb = DatasetBuilder(ml_ds)

    rates = ml_ds.interactions().pandas(attributes=[], ids=True)
    samp_df = rates.sample(1000, replace=False, random_state=rng, ignore_index=True)
    samp = pa.Table.from_pandas(samp_df, preserve_index=False)

    dsb.filter_interactions("rating", remove=samp)

    ds = dsb.build()
    assert ds.interaction_count == ml_ds.interaction_count - 1000

    dsb2 = DatasetBuilder(ds)
    dsb2.add_interactions("rating", samp)

    ds2 = dsb2.build()
    assert ds2.interaction_count == ml_ds.interaction_count



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from enum import Enum
from hashlib import blake2b
from typing import NamedTuple
from uuid import UUID

from pydantic import BaseModel


class MsgAuthenticator:
    _authkey: bytes

    def __init__(self, key: bytes):
        self._authkey = key

    def hash_message(self, channel: LogChannel | bytes, name: bytes, data: bytes) -> bytes:
        mb = blake2b(key=self._authkey)
        if isinstance(channel, bytes):
            mb.update(channel)
        else:
            mb.update(channel.value)
        mb.update(name)
        mb.update(data)
        return mb.digest()

    def verify_message(
        self, channel: LogChannel | bytes, name: bytes, data: bytes, hash: bytes
    ) -> bool:
        return self.hash_message(channel, name, data) == hash


class LogChannel(Enum):
    STDLIB = b"stdlib"
    STRUCTLOG = b"structlog"
    TASKS = b"lenskit.logging.tasks"
    PROGRESS = b"lenskit.logging.progress"
    RECORD = b"lenskit.logging.record"


class ProgressField(NamedTuple):
    value: int | float | str | None
    format: str | None = None


class ProgressMessage(BaseModel):
    progress_id: UUID
    label: str
    total: int | float | None
    completed: int | float | None
    fields: dict[str, ProgressField] = {}
    finished: bool = False



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from uuid import uuid4

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import torch

from pytest import mark

from lenskit.data import ItemList, Vocabulary


@mark.benchmark(group="take")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_itemlist_take(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    items = ItemList(np.arange(size), scores=scores)
    ranks = np.argsort(scores)

    def take():
        _il = items[ranks]

    benchmark(take)


@mark.benchmark(group="take")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_arrow_table_take(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    items = ItemList(np.arange(size), scores=scores)
    table = items.to_arrow()
    ranks = np.argsort(scores)
    ranks = pa.array(ranks, type=pa.int32())

    def take():
        _il = table.take(ranks)

    benchmark(take)


@mark.benchmark(group="take")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_arrow_struct_take(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    items = ItemList(np.arange(size), scores=scores)
    table = items.to_arrow(type="array")
    ranks = np.argsort(scores)
    ranks = pa.array(ranks, type=pa.int32())

    def take():
        _il = table.take(ranks)

    benchmark(take)


@mark.benchmark(group="take")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_arrow_table_take_cvt(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    items = ItemList(np.arange(size), scores=scores)
    table = items.to_arrow()
    ranks = np.argsort(scores)

    def take():
        aranks = pa.array(ranks, type=pa.int32())
        _il = table.take(aranks)

    benchmark(take)


@mark.benchmark(group="take")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_pandas_take(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    items = ItemList(np.arange(size), scores=scores)
    df = items.to_df()
    ranks = np.argsort(scores)

    def take():
        _il = df.iloc[ranks]

    benchmark(take)


@mark.benchmark(group="take")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_numpy_take(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    items = ItemList(np.arange(size), scores=scores)
    ranks = np.argsort(scores)

    def take():
        _ids = items.ids()[ranks]
        _scores = items.ids()[ranks]

    benchmark(take)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from pytest import approx, mark

from lenskit.data import ItemList
from lenskit.metrics import call_metric
from lenskit.metrics.ranking import DCG
from lenskit.metrics.ranking._dcg import array_dcg, fixed_dcg


@mark.parametrize("method", ["graded", "binary"])
def test_array_dcg_empty(method):
    "empty should be zero"
    assert array_dcg(np.array([]), graded=method == "graded") == approx(0)


@mark.parametrize("method", ["graded", "binary"])
def test_array_dcg_zeros(method):
    assert array_dcg(np.zeros(10), graded=method == "graded") == approx(0)


@mark.parametrize("method", ["graded", "binary"])
def test_array_dcg_single(method):
    "a single element should be scored at the right place"
    val = 0.5 if method == "graded" else 1.0
    assert array_dcg(np.array([0.5]), graded=method == "graded") == approx(val)
    assert array_dcg(np.array([0, 0.5]), graded=method == "graded") == approx(val)
    assert array_dcg(np.array([0, 0, 0.5]), graded=method == "graded") == approx(val / np.log2(3))
    assert array_dcg(np.array([0, 0, 0.5, 0]), graded=method == "graded") == approx(
        val / np.log2(3)
    )


def test_array_dcg_mult():
    "multiple elements should score correctly"
    assert array_dcg(np.array([np.e, np.pi])) == approx(np.e + np.pi)
    assert array_dcg(np.array([np.e, 0, 0, np.pi])) == approx(np.e + np.pi / np.log2(4))


@mark.parametrize("method", ["graded", "binary"])
def test_array_dcg_empty2(method):
    "empty should be zero"
    assert array_dcg(np.array([]), graded=method == "graded") == approx(0)


@mark.parametrize("method", ["graded", "binary"])
def test_array_dcg_zeros2(method):
    assert array_dcg(np.zeros(10), graded=method == "graded") == approx(0)


def test_array_dcg_nan():
    "NANs should be 0"
    assert array_dcg(np.array([np.nan, 0.5])) == approx(0.5)


def test_dcg_empty():
    recs = ItemList(ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth) == approx(0.0)


def test_dcg_no_match():
    recs = ItemList([4], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth) == approx(0.0)


def test_dcg_perfect():
    recs = ItemList([2, 3, 1], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth) == approx(np.sum(np.reciprocal(np.log2([2, 2, 3]))))


def test_dcg_perfect_k_short():
    recs = ItemList([2, 3, 1], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth, k=2) == approx(2.0)
    assert call_metric(DCG, recs[:2], truth, k=2) == approx(2.0)


def test_dcg_shorter_not_best():
    recs = ItemList([1, 2], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth) == approx(fixed_dcg(2))
    assert call_metric(DCG, recs, truth, k=2) == approx(2.0)
    assert call_metric(DCG, recs, truth, gain="rating") == approx(array_dcg(np.array([3.0, 5.0])))


def test_dcg_perfect_k():
    recs = ItemList([2, 3], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth, k=2) == approx(2.0)


def test_dcg_perfect_k_norate():
    recs = ItemList([1, 3], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth, k=2) == approx(2.0)


def test_dcg_almost_perfect_k_gain():
    recs = ItemList([1, 3], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(DCG, recs, truth, k=2, gain="rating") == approx(
        array_dcg(np.array([3.0, 4.0]))
    )



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
from os import fspath
from pathlib import Path

from click.testing import CliRunner

from lenskit.cli import lenskit
from lenskit.data import Dataset, ItemListCollection


def test_data_subset(tmpdir: Path, ml_ds: Dataset):
    ds_path = tmpdir / "ml-data"
    ml_ds.save(ds_path)

    out_path = tmpdir / "ml.subset.parquet"

    runner = CliRunner()
    result = runner.invoke(
        lenskit,
        ["data", "subset", "--sample-rows=100", "--item-lists", fspath(ds_path), fspath(out_path)],
    )

    assert result.exit_code == 0
    assert out_path.exists()

    items = ItemListCollection.load_parquet(out_path)
    assert len(items) == 100



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Literal

import numpy as np
from scipy.special import softmax

from lenskit.data import ItemList
from lenskit.data.query import QueryInput, RecQuery
from lenskit.data.types import NPVector
from lenskit.pipeline import Component
from lenskit.random import DerivableSeed, RNGFactory, derivable_rng


@dataclass
class StochasticTopNConfig:
    """
    Configuration for :class:`StochasticTopNRanker`.
    """

    n: int | None = None
    """
    The number of items to select. -1 or ``None`` to return all scored items.
    """

    rng: DerivableSeed = None
    """
    Random number generator configuration.
    """

    transform: Literal["softmax", "linear"] | None = "softmax"
    """
    Transformation to convert scores into ranking probabilities.

    softmax
        Use the softmax of the item scores as the selection probabilities.
    linear
        Linearly re-scale item scores to be selection probabilities. This
        equivalent to min-max scaling the scores, then re-scaling to sum
        to 1.
    ``None``
        No transformation, except negative scores are clamped to (almost)
        zero.  Not recommended unless your item scorer emits multinomial
        probabilities.
    """
    scale: float = 1.0
    """
    Scalar multiplier to apply to scores prior to transformation.  This is
    equivalent to the :math:`\\beta` parameter for parameterized softmax
    transformation.  Larger values will decrease the entropy of the sampled
    rankings.
    """


class StochasticTopNRanker(Component[ItemList]):
    """
    Stochastic top-N ranking with optional weight transformation.

    This uses the exponential sampling method, a more efficient approximation of
    Plackett-Luce sampling than even the Gumbell trick, as documented by `Tim
    Vieira`_.  It expects a scored list of input items, and samples ``n`` items,
    with selection probabilities proportional to their scores.  Scores can be
    optionally rescaled (inverse temperature) and transformed (e.g. softmax).

    .. note::

        When no transformation is used, negative scores are still clamped to
        (approximately) zero.

    .. _Tim Vieira: https://timvieira.github.io/blog/post/2019/09/16/algorithms-for-sampling-without-replacement/

    Stability:
        Caller
    """

    config: StochasticTopNConfig
    "Stochastic ranker configuration."

    _rng_factory: RNGFactory

    def __init__(self, config: StochasticTopNConfig | None = None, **kwargs):
        super().__init__(config, **kwargs)
        self._rng_factory = derivable_rng(self.config.rng)

    def __call__(
        self,
        items: ItemList,
        query: QueryInput | None = None,
        n: int | None = None,
        *,
        include_weights: bool = False,
    ) -> ItemList:
        query = RecQuery.create(query)
        rng = self._rng_factory(query)

        scores = items.scores()
        if scores is None:
            raise ValueError("item list must have scores")

        valid_mask = np.isfinite(scores)
        valid_items = items[valid_mask]
        N = len(valid_items)
        if N == 0:
            return ItemList(valid_items, ordered=True)

        if n is None or n < 0:
            n = self.config.n or -1

        keys = self._compute_keys(scores[valid_mask], rng)

        if include_weights:
            valid_items = ItemList(valid_items, weight=keys)
        return valid_items.top_n(n, scores=keys)

    def _compute_keys(self, scores: NPVector, rng: np.random.Generator) -> NPVector:
        "Compute sort keys for sampled ranking."
        # scale the scores — with softmax, this is the equivalent of β.
        # see: https://en.wikipedia.org/wiki/Softmax_function
        scores = scores * self.config.scale
        match self.config.transform:
            case "linear":
                lb = np.min(scores).item()
                ub = np.max(scores).item()
                r = ub - lb
                # scale weights twice to reduce risk of floating-point error
                scaled = scores - lb
                weights = None
                if r > 0:
                    scaled /= r
                    tot = np.sum(scaled)
                    if tot > 0:
                        weights = scaled / tot

                if weights is None:
                    weights = np.ones_like(scores) / len(scores)
            case "softmax":
                weights = softmax(scores)
                assert np.all(np.isfinite(weights))
                assert np.all(weights >= 0)
                assert np.all(weights <= 1)
            case None:
                weights = scores

        # positive instead of negative, because we take top-N instead of bottom
        keys = np.log(rng.uniform(0, 1, len(scores)))
        assert np.all(np.isfinite(keys))
        assert np.all(keys < 0)
        # smoooth very small weights to avoid divide-by-zero
        keys /= np.maximum(weights, np.finfo("f4").smallest_normal)
        assert np.all(np.isfinite(keys))

        return keys



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Sparse matrix utility functions.
"""

# pyright: basic
from __future__ import annotations

import logging
from typing import Literal

import scipy.sparse as sps
import torch

from lenskit.torch import safe_tensor

_log = logging.getLogger(__name__)


def torch_sparse_from_scipy(
    M: sps.coo_array, layout: Literal["csr", "coo", "csc"] = "coo"
) -> torch.Tensor:
    """
    Convert a SciPy :class:`sps.coo_array` into a torch sparse tensor.

    Stability:
        Internal
    """
    ris = safe_tensor(M.row)
    cis = safe_tensor(M.col)
    vs = safe_tensor(M.data)
    indices = torch.stack([ris, cis])
    assert indices.shape == (2, M.nnz)
    T = torch.sparse_coo_tensor(indices, vs, size=M.shape)
    assert T.shape == M.shape

    match layout:
        case "csr":
            return T.to_sparse_csr()
        case "csc":
            return T.to_sparse_csc()
        case "coo":
            return T.coalesce()
        case _:
            raise ValueError(f"invalid layout {layout}")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
from os import fspath
from pathlib import Path

from click.testing import CliRunner

from lenskit.cli import lenskit
from lenskit.data import Dataset, ItemListCollection


def test_data_split(tmpdir: Path, ml_ds: Dataset):
    ds_path = tmpdir / "ml-data"
    ml_ds.save(ds_path)

    train_path = tmpdir / "ml-data.train"
    test_path = tmpdir / "ml-data.test.parquet"

    runner = CliRunner()
    result = runner.invoke(
        lenskit,
        ["data", "split", "--fraction", "0.2", fspath(ds_path)],
    )

    assert result.exit_code == 0
    assert train_path.exists()
    assert test_path.exists()

    trds = Dataset.load(train_path)
    print(trds)
    items = ItemListCollection.load_parquet(test_path)
    test_n = sum(len(il) for il in items.lists())
    tgt_n = int(ml_ds.interaction_count * 0.2)
    assert tgt_n - 1 <= test_n <= tgt_n + 1

    assert trds.interaction_count + test_n == ml_ds.interaction_count


def test_data_split_time(tmpdir: Path, ml_ds: Dataset):
    ds_path = tmpdir / "ml-data"
    ml_ds.save(ds_path)

    train_path = tmpdir / "ml-data.train"
    test_path = tmpdir / "ml-data.test.parquet"

    runner = CliRunner()
    result = runner.invoke(
        lenskit,
        ["data", "split", "--date", "2015-01-01", fspath(ds_path)],
    )

    assert result.exit_code == 0
    assert train_path.exists()
    assert test_path.exists()

    trds = Dataset.load(train_path)
    print(trds)
    items = ItemListCollection.load_parquet(test_path)
    test_n = sum(len(il) for il in items.lists())

    assert trds.interaction_count + test_n == ml_ds.interaction_count


def test_data_out_dir(tmpdir: Path, ml_ds: Dataset):
    ds_path = tmpdir / "ml-data"
    ml_ds.save(ds_path)

    out_dir = tmpdir / "split"
    train_path = out_dir / "train"
    test_path = out_dir / "test.parquet"

    runner = CliRunner()
    result = runner.invoke(
        lenskit,
        ["data", "split", "--fraction", "0.2", "-d", fspath(out_dir), fspath(ds_path)],
    )

    assert result.exit_code == 0
    assert train_path.exists()
    assert test_path.exists()

    trds = Dataset.load(train_path)
    print(trds)
    items = ItemListCollection.load_parquet(test_path)
    test_n = sum(len(il) for il in items.lists())
    tgt_n = int(ml_ds.interaction_count * 0.2)
    assert tgt_n - 1 <= test_n <= tgt_n + 1

    assert trds.interaction_count + test_n == ml_ds.interaction_count



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for the Dataset class.
"""

import numpy as np
import pandas as pd

from lenskit.data import Dataset
from lenskit.testing import ml_ds, ml_ratings  # noqa: F401


def test_pandas_log_defaults(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    int_df = ml_ds.interaction_table(format="pandas")
    assert isinstance(int_df, pd.DataFrame)
    # we should have exactly the 4 expected columns
    assert len(int_df.columns) == 4
    assert "user_num" in int_df.columns
    assert "item_num" in int_df.columns
    assert "rating" in int_df.columns
    assert "timestamp" in int_df.columns

    # the interact
    int_df = int_df.sort_values(["user_num", "item_num"])
    uids = ml_ds.users.ids(int_df["user_num"])
    iids = ml_ds.items.ids(int_df["item_num"])

    ml_df = ml_ratings.sort_values(["user_id", "item_id"])
    assert np.all(uids == ml_df["user_id"])
    assert np.all(iids == ml_df["item_id"])
    assert np.all(int_df["rating"] == ml_df["rating"])
    assert np.all(int_df["timestamp"] == ml_df["timestamp"])

    # and the total length
    assert len(int_df) == len(ml_ratings)


def test_pandas_log_ids(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    int_df = ml_ds.interaction_table(format="pandas", original_ids=True)
    assert isinstance(int_df, pd.DataFrame)
    # we should have exactly the 4 expected columns
    assert len(int_df.columns) == 4
    assert "user_id" in int_df.columns
    assert "item_id" in int_df.columns
    assert "rating" in int_df.columns
    assert "timestamp" in int_df.columns

    # the interact
    int_df = int_df.sort_values(["user_id", "item_id"])

    ml_df = ml_ratings.sort_values(["user_id", "item_id"])
    assert np.all(int_df["user_id"] == ml_df["user_id"])
    assert np.all(int_df["item_id"] == ml_df["item_id"])
    assert np.all(int_df["rating"] == ml_df["rating"])
    assert np.all(int_df["timestamp"] == ml_df["timestamp"])

    # and the total length
    assert len(int_df) == len(ml_ratings)


def test_pandas_log_no_ts(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    int_df = ml_ds.interaction_table(format="pandas", fields=["rating"])
    assert isinstance(int_df, pd.DataFrame)
    # we should have exactly the 4 expected columns
    assert len(int_df.columns) == 3
    assert "user_num" in int_df.columns
    assert "item_num" in int_df.columns
    assert "rating" in int_df.columns

    # the interact
    int_df = int_df.sort_values(["user_num", "item_num"])
    uids = ml_ds.users.ids(int_df["user_num"])
    iids = ml_ds.items.ids(int_df["item_num"])

    ml_df = ml_ratings.sort_values(["user_id", "item_id"])
    assert np.all(uids == ml_df["user_id"])
    assert np.all(iids == ml_df["item_id"])
    assert np.all(int_df["rating"] == ml_df["rating"])

    # and the total length
    assert len(int_df) == len(ml_ratings)


def test_numpy_log_defaults(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_table(format="numpy")
    assert isinstance(log, dict)
    assert log["rating"] is not None
    assert log["timestamp"] is not None

    # and the total length
    assert len(log["user_num"]) == len(ml_ratings)
    assert len(log["item_num"]) == len(ml_ratings)
    assert len(log["rating"]) == len(ml_ratings)
    assert len(log["timestamp"]) == len(ml_ratings)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic set statistics.
"""

import numpy as np

from lenskit.data.items import ItemList

from ._base import ListMetric, Metric


class ListLength(ListMetric):
    """
    Report the length of the output (recommendation list or predictions).

    Stability:
        Caller
    """

    label = "N"  # type: ignore

    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        return len(recs)

    def summarize(self, values):
        return float(np.mean(values))


class TestItemCount(Metric):
    """
    Report the number of test items.

    Stability:
        Caller
    """

    label = "TestItemCount"  # type: ignore

    def __call__(self, recs: ItemList, test: ItemList) -> float:
        return len(test)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from lenskit.metrics.bulk import RunAnalysis
from lenskit.metrics.ranking._gini import ExposureGini, ListGini
from lenskit.testing import DemoRecs, demo_recs, pop_recs


def test_list_gini(pop_recs: DemoRecs):
    rla = RunAnalysis()
    rla.add_metric(ListGini(items=pop_recs.split.train))

    result = rla.measure(pop_recs.recommendations, pop_recs.split.test)
    rm = result.global_metrics()
    assert "ListGini" in rm.index
    # most-popular is very high Gini
    assert rm["ListGini"] < 1
    assert rm["ListGini"] > 0.90


def test_exposure_gini(pop_recs: DemoRecs):
    rla = RunAnalysis()
    rla.add_metric(ExposureGini(items=pop_recs.split.train))

    result = rla.measure(pop_recs.recommendations, pop_recs.split.test)
    rm = result.global_metrics()
    assert "ExposureGini" in rm.index
    # most-popular is very high Gini
    assert rm["ExposureGini"] < 1
    assert rm["ExposureGini"] > 0.95


def test_gini_changes(pop_recs: DemoRecs, demo_recs: DemoRecs):
    "test that randomization improves gini"

    rla = RunAnalysis()
    rla.add_metric(ListGini(items=pop_recs.split.train))
    rla.add_metric(ExposureGini(items=pop_recs.split.train))

    pop_result = rla.measure(pop_recs.recommendations, pop_recs.split.test)
    samp_result = rla.measure(demo_recs.recommendations, demo_recs.split.test)

    pop_rm = pop_result.global_metrics()
    samp_rm = samp_result.global_metrics()

    # most-popular should higher gini on both versions
    assert pop_rm["ExposureGini"] > samp_rm["ExposureGini"]
    assert pop_rm["ListGini"] > samp_rm["ListGini"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from datetime import datetime

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import approx, mark, raises, warns

from lenskit.data import Dataset, DatasetBuilder
from lenskit.data.schema import AllowableTroolean
from lenskit.diagnostics import DataError, DataWarning


def test_add_interactions_insert_ids_df():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {"user_id": ["a", "a", "b", "c", "c", "c"], "item_id": ["x", "y", "z", "x", "y", "z"]}
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ecs = dsb.entity_classes()
    assert set(ecs.keys()) == {"user", "item"}

    rcs = dsb.relationship_classes()
    assert set(rcs.keys()) == {"click"}
    print(dsb.schema.model_dump_json(indent=2))

    ucls = ecs["user"]
    assert ucls.id_type == "str"

    ds = dsb.build()
    assert ds.user_count == 3
    assert ds.item_count == 3
    assert np.all(ds.users.ids() == ["a", "b", "c"])
    assert np.all(ds.items.ids() == ["x", "y", "z"])

    istats = ds.item_stats()
    print(istats)
    assert np.all(istats["user_count"] == 2)

    log = ds.interaction_table(format="pandas", original_ids=True)
    assert isinstance(log, pd.DataFrame)
    print(log)
    assert all(log.columns == ["user_id", "item_id"])
    assert len(log) == 6

    mat = ds.interaction_matrix(format="structure")
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])


def test_add_interactions_forbidden_attribute_name():
    dsb = DatasetBuilder()
    with raises(ValueError, match="invalid"):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["a", "a", "b", "c", "c", "c"],
                    "item_id": ["x", "y", "z", "x", "y", "z"],
                    "_count": np.arange(1, 7),
                }
            ),
            entities=["user", "item"],
            missing="insert",
        )

    with raises(ValueError, match="invalid"):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["a", "a", "b", "c", "c", "c"],
                    "item_id": ["x", "y", "z", "x", "y", "z"],
                    "count_id": np.arange(1, 7),
                }
            ),
            entities=["user", "item"],
            missing="insert",
        )


def test_add_interactions_table():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pa.table(
            {
                "user_id": ["a", "a", "b", "c", "c", "c"],
                "item_id": ["x", "y", "z", "x", "y", "z"],
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ecs = dsb.entity_classes()
    assert set(ecs.keys()) == {"user", "item"}

    rcs = dsb.relationship_classes()
    assert set(rcs.keys()) == {"click"}

    ucls = ecs["user"]
    assert ucls.id_type == "str"

    ds = dsb.build()
    assert ds.user_count == 3
    assert ds.item_count == 3
    assert np.all(ds.users.ids() == ["a", "b", "c"])
    assert np.all(ds.items.ids() == ["x", "y", "z"])

    istats = ds.item_stats()
    assert np.all(istats["user_count"] == 2)

    log = ds.interaction_table(format="pandas", original_ids=True)
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_id", "item_id"])
    assert len(log) == 6

    mat = ds.interaction_matrix(format="structure")
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])


def test_add_interactions_count_method():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c"],
                "item_id": ["x", "y", "z", "x", "y", "z"],
                "count": np.arange(1, 7),
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()
    log = ds.interactions().count()
    assert log == 21


def test_add_interactions_count_method_with_null():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c"],
                "item_id": ["x", "y", "z", "x", "y", "z"],
                "count": [1, 2, None, 4, 5, 6],
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()
    log = ds.interactions().count()
    assert log == 19


def test_add_interactions_error_bad_ids():
    dsb = DatasetBuilder()
    dsb.add_entities("user", ["a", "b", "c"])
    dsb.add_entities("item", ["z", "x", "y"])

    # fail with missing entities
    with raises(DataError, match="unknown"):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["a", "a", "b", "c", "c", "c"],
                    "item_id": ["x", "y", "z", "x", "y", "w"],
                }
            ),
            entities=["user", "item"],
            missing="error",
        )

    # but using the correct ones passes
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c"],
                "item_id": ["x", "y", "z", "x", "y", "z"],
            }
        ),
        entities=["user", "item"],
        missing="error",
    )

    ds = dsb.build()
    assert ds.user_count == 3
    assert ds.item_count == 3
    assert np.all(ds.users.ids() == ["a", "b", "c"])
    assert np.all(ds.items.ids() == ["x", "y", "z"])

    log = ds.interaction_table(format="pandas", original_ids=True)
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_id", "item_id"])
    assert len(log) == 6

    mat = ds.interaction_matrix(format="structure")
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])

    assert len(ds.user_row("a")) == 2
    assert len(ds.user_row("b")) == 1
    assert len(ds.user_row("c")) == 3


def test_add_interactions_filter_bad_ids():
    dsb = DatasetBuilder()
    dsb.add_entities("user", ["a", "b", "c"])
    dsb.add_entities("item", ["z", "x", "y"])

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c"],
                "item_id": ["x", "y", "z", "x", "w", "z"],
            }
        ),
        entities=["user", "item"],
        missing="filter",
    )

    ds = dsb.build()
    assert ds.user_count == 3
    assert ds.item_count == 3
    assert np.all(ds.users.ids() == ["a", "b", "c"])
    assert np.all(ds.items.ids() == ["x", "y", "z"])

    log = ds.interaction_table(format="pandas", original_ids=True)
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_id", "item_id"])
    assert len(log) == 5

    mat = ds.interaction_matrix(format="structure")
    assert mat.nnz == 5
    assert np.all(mat.rowptrs == [0, 2, 3, 5])

    assert len(ds.user_row("a")) == 2
    assert len(ds.user_row("b")) == 1
    assert len(ds.user_row("c")) == 2


def test_add_interactions_forbidden_repeat():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=False)
    with raises(DataError, match="repeated"):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                    "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                    "timestamp": np.arange(1, 8) * 10,
                }
            ),
            missing="insert",
        )
        dsb.build()


def test_add_multiple_interactions_forbidden_repeat():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=False)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c"],
                "item_id": ["x", "y", "z", "x"],
                "timestamp": np.arange(1, 5) * 10,
            }
        ),
        missing="insert",
    )
    with raises(DataError, match="repeated"):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["c", "c", "b"],
                    "item_id": ["y", "z", "z"],
                    "timestamp": np.arange(5, 8) * 10,
                }
            ),
            missing="insert",
        )
        dsb.build()


def test_add_auto_entities():
    dsb = DatasetBuilder()

    with warns(DataWarning, match="specified"):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                    "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                    "timestamp": np.arange(1, 8) * 10,
                }
            ),
            missing="insert",
        )

    dsb.build()
    rsc = dsb.relationship_classes()["click"]
    assert rsc.repeats == AllowableTroolean.PRESENT
    assert rsc.entity_class_names == ["user", "item"]


def test_add_ratings(ml_ratings: pd.DataFrame):
    dsb = DatasetBuilder()
    dsb.add_interactions("rating", ml_ratings, entities=["user", "item"], missing="insert")

    rsc = dsb.relationship_classes()["rating"]
    assert rsc.entity_class_names == ["user", "item"]

    db = dsb.build()
    assert db.user_count == ml_ratings["user_id"].nunique()
    assert db.item_count == ml_ratings["item_id"].nunique()
    assert db.interaction_count == len(ml_ratings)

    ldf = db.interaction_table(format="pandas", original_ids=True)
    assert "rating" in ldf.columns
    assert ldf["rating"].mean() == approx(ml_ratings["rating"].mean())
    assert ldf["timestamp"].max() == ml_ratings["timestamp"].max()


def test_add_ratings_batched(ml_ratings: pd.DataFrame):
    dsb = DatasetBuilder()

    for bstart in range(0, len(ml_ratings), 1000):
        bend = min(len(ml_ratings), bstart + 1000)
        df = ml_ratings.iloc[bstart:bend]
        dsb.add_interactions("rating", df, entities=["user", "item"], missing="insert")

    rsc = dsb.relationship_classes()["rating"]
    assert rsc.entity_class_names == ["user", "item"]

    db = dsb.build()
    assert db.user_count == ml_ratings["user_id"].nunique()
    assert db.item_count == ml_ratings["item_id"].nunique()
    assert db.interaction_count == len(ml_ratings)

    ldf = db.interaction_table(format="pandas", original_ids=True)
    assert "rating" in ldf.columns
    assert ldf["rating"].mean() == approx(ml_ratings["rating"].mean())
    assert ldf["timestamp"].max() == ml_ratings["timestamp"].max()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import click

from lenskit import lenskit_config
from lenskit.logging import get_logger, stdout_console

_log = get_logger(__name__)
_gh_out: Path | None = None


@click.command("config")
def config():
    """
    Inspect LensKit configuration.
    """
    console = stdout_console()
    config = lenskit_config()
    console.print_json(config.model_dump_json(indent=2))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
import pickle
from os import fspath
from pathlib import Path

from click.testing import CliRunner
from xopen import xopen

from lenskit.als import BiasedMFScorer
from lenskit.cli import lenskit
from lenskit.data import Dataset
from lenskit.pipeline.nodes import ComponentInstanceNode


def test_train_config(tmpdir: Path, ml_ds: Dataset):
    ds_path = tmpdir / "ml-data"
    ml_ds.save(ds_path)

    out_file = tmpdir / "als.pkl.gz"

    runner = CliRunner()
    result = runner.invoke(
        lenskit,
        [
            "train",
            "--config",
            "pipelines/als-explicit.toml",
            "-o",
            fspath(out_file),
            fspath(ds_path),
        ],
    )

    assert result.exit_code == 0
    assert out_file.exists()

    with xopen(out_file, "rb") as pf:
        pipe = pickle.load(pf)

    node = pipe.node("scorer")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, BiasedMFScorer)
